ID_VARS = set(["KeyGroup", "ParamGroup", "FilePath"])
NON_KEY_ENTITIES = set(["subject", "session", "extension"])
# Multi-dimensional keys SliceTiming
IMAGING_PARAMS = set([
    "ParallelReductionFactorInPlane", "ParallelAcquisitionTechnique",
    "ParallelAcquisitionTechnique", "PartialFourier", "PhaseEncodingDirection",
    "EffectiveEchoSpacing", "TotalReadoutTime", "EchoTime",
    "SliceEncodingDirection", "DwellTime", "FlipAngle",
    "MultibandAccelerationFactor", "RepetitionTime",
    "VolumeTiming", "NumberOfVolumesDiscardedByScanner",
    "NumberOfVolumesDiscardedByUser", "Obliquity", "VoxelSizeDim1",
    "VoxelSizeDim2", "VoxelSizeDim3", "Dim1Size", "Dim2Size", "Dim3Size",
    "NumVolumes"])



#!/usr/bin/env python

"""Tests for `cubids` package."""
import sys
sys.path.append("..")
import shutil
from copy import deepcopy
import hashlib
import json
from pathlib import Path
from pkg_resources import resource_filename as pkgrf
import pytest
from cubids import CuBIDS
from cubids.validator import (build_validator_call,
                       run_validator, parse_validator_output)
from cubids.metadata_merge import (
    merge_without_overwrite, merge_json_into_json)
from math import nan
import subprocess
import csv
import os
import filecmp
import nibabel as nb
import numpy as np
import pandas as pd
import subprocess

TEST_DATA = pkgrf("cubids", "testdata")

COMPLETE_KEY_GROUPS = [
    'acquisition-HASC55AP_datatype-dwi_suffix-dwi',
    'acquisition-v4_datatype-fmap_fmap-magnitude1_suffix-magnitude1',
    'acquisition-v4_datatype-fmap_fmap-magnitude2_suffix-magnitude2',
    'acquisition-v4_datatype-fmap_fmap-phasediff_suffix-phasediff',
    'datatype-anat_suffix-T1w',
    'datatype-fmap_direction-PA_fmap-epi_suffix-epi',
    'datatype-func_suffix-bold_task-rest']


def get_data(tmp_path):
    """Copy testing data to a local directory"""
    data_root = tmp_path / "testdata"
    shutil.copytree(TEST_DATA, str(data_root))
    return data_root


def test_ok_json_merge(tmp_path):
    data_root = get_data(tmp_path)

    # Test that a successful merge can happen
    dest_json = data_root / "inconsistent" / "sub-02" / \
        "ses-phdiff" / "dwi" / "sub-02_ses-phdiff_acq-HASC55AP_dwi.json"
    orig_dest_json_content = _get_json_string(dest_json)
    source_json = data_root / "inconsistent" / "sub-03" / \
        "ses-phdiff" / "dwi" / "sub-03_ses-phdiff_acq-HASC55AP_dwi.json"

    merge_return = merge_json_into_json(source_json, dest_json)
    assert merge_return == 0
    assert not _get_json_string(dest_json) == orig_dest_json_content


def test_ok_json_merge_cli(tmp_path):
    data_root = get_data(tmp_path)

    # Test that a successful merge can happen
    dest_json = data_root / "inconsistent" / "sub-02" / \
        "ses-phdiff" / "dwi" / "sub-02_ses-phdiff_acq-HASC55AP_dwi.json"
    orig_dest_json_content = _get_json_string(dest_json)
    source_json = data_root / "inconsistent" / "sub-03" / \
        "ses-phdiff" / "dwi" / "sub-03_ses-phdiff_acq-HASC55AP_dwi.json"

    merge_proc = subprocess.run(
        ['bids-sidecar-merge', str(source_json), str(dest_json)])
    assert merge_proc.returncode == 0
    assert not _get_json_string(dest_json) == orig_dest_json_content

def test_get_param_groups(tmp_path):
    data_root = get_data(tmp_path)
    bod = CuBIDS(data_root / "inconsistent", use_datalad=True)
    csv_prefix = str(tmp_path / "csvs")
    key_groups = bod.get_key_groups()
    bod._cache_fieldmaps()

    for key_group in key_groups:
        ret = bod.get_param_groups_from_key_group(key_group)
        assert sum(ret[1].Counts) == ret[1].loc[0, 'KeyGroupCount']

def test_copy_exemplars(tmp_path):
    data_root = get_data(tmp_path)
    bod = CuBIDS(data_root / "complete", use_datalad=True)
    csv_prefix = str(tmp_path / "csvs")
    bod.get_CSVs(csv_prefix)
    acq_group_csv = csv_prefix + "_AcqGrouping.csv"
    print("ACQ GROUP PATH: ", acq_group_csv)
    exemplars_dir = str(tmp_path / "exemplars")
    print('EXEMPLARS DIR: ', exemplars_dir)
    df = pd.read_csv(acq_group_csv)

    bod.copy_exemplars(exemplars_dir, acq_group_csv, min_group_size=1)

    # check exemplar dir got created and has the correct number of subs
    cntr = 0
    for path in Path(exemplars_dir).glob("sub-*"):
        cntr += 1
    assert cntr == len(df.drop_duplicates(subset=["AcqGroup"]))

    # check that dataset_description.json got added
    assert Path(exemplars_dir + '/dataset_description.json').exists()

def test_purge_no_datalad(tmp_path):
    data_root = get_data(tmp_path)
    scans = []
    scan_name = data_root / "complete" / "sub-03" / "ses-phdiff" \
        / "func" / "sub-03_ses-phdiff_task-rest_bold.nii.gz"
    json_name = data_root / "complete" / "sub-03" / "ses-phdiff" \
        / "func" / "sub-03_ses-phdiff_task-rest_bold.json"
    scans.append(scan_name)
    scans.append(data_root / "complete" / "sub-01" / "ses-phdiff/dwi/sub-01_ses-phdiff_acq-HASC55AP_dwi.nii.gz")

    # create and save .txt with list of scans
    purge_path = str(tmp_path / "purge_scans.txt")
    with open(purge_path, 'w') as filehandle:
        for listitem in scans:
            filehandle.write('%s\n' % listitem)
    bod = CuBIDS(data_root / "complete", use_datalad=False)

    assert Path(scan_name).exists()
    assert Path(json_name).exists()

    # Check that IntendedFor purge worked
    with open(str(data_root / "complete" / "sub-01" / "ses-phdiff" / "fmap" / "sub-01_ses-phdiff_acq-v4_phasediff.json")) as f:
        j_dict = json.load(f)

    assert "ses-phdiff/dwi/sub-01_ses-phdiff_acq-HASC55AP_dwi.nii.gz" in j_dict.values()
    assert isinstance(j_dict['IntendedFor'], str)
    # PURGE
    bod.purge(purge_path)

    with open(str(data_root / "complete" / "sub-01" / "ses-phdiff" / "fmap" / "sub-01_ses-phdiff_acq-v4_phasediff.json")) as f:
        purged_dict = json.load(f)

    assert not Path(scan_name).exists()
    assert not Path(json_name).exists()
    assert "ses-phdiff/dwi/sub-01_ses-phdiff_acq-HASC55AP_dwi.nii.gz" not in purged_dict.values()
    assert isinstance(purged_dict['IntendedFor'], list)
    assert purged_dict['IntendedFor'] == []

def test_purge(tmp_path):
    data_root = get_data(tmp_path)
    scans = []
    scan_name = data_root / "complete" / "sub-03" / "ses-phdiff" \
        / "func" / "sub-03_ses-phdiff_task-rest_bold.nii.gz"
    json_name = data_root / "complete" / "sub-03" / "ses-phdiff" \
        / "func" / "sub-03_ses-phdiff_task-rest_bold.json"
    scans.append(scan_name)
    purge_path = str(tmp_path / "purge_scans.txt")

    with open(purge_path, 'w') as filehandle:
        for listitem in scans:
            filehandle.write('%s\n' % listitem)
    bod = CuBIDS(data_root / "complete", use_datalad=True)
    bod.datalad_save()

    assert bod.is_datalad_clean()
    assert Path(scan_name).exists()
    assert Path(json_name).exists()

    # create and save .txt with list of scans
    bod.purge(purge_path)

    assert not Path(scan_name).exists()
    assert not Path(json_name).exists()


def test_bad_json_merge(tmp_path):
    data_root = get_data(tmp_path)

    # Test that a successful merge can happen
    dest_json = data_root / "inconsistent" / "sub-02" / \
        "ses-phdiff" / "dwi" / "sub-02_ses-phdiff_acq-HASC55AP_dwi.json"
    orig_dest_json_content = _get_json_string(dest_json)
    invalid_source_json = data_root / "inconsistent" / "sub-01" / \
        "ses-phdiff" / "dwi" / "sub-01_ses-phdiff_acq-HASC55AP_dwi.json"

    assert merge_json_into_json(invalid_source_json, dest_json) > 0
    assert _get_json_string(dest_json) == orig_dest_json_content


def test_bad_json_merge_cli(tmp_path):
    data_root = get_data(tmp_path)

    # Test that a successful merge can happen
    dest_json = data_root / "inconsistent" / "sub-02" / \
        "ses-phdiff" / "dwi" / "sub-02_ses-phdiff_acq-HASC55AP_dwi.json"
    orig_dest_json_content = _get_json_string(dest_json)
    invalid_source_json = data_root / "inconsistent" / "sub-01" / \
        "ses-phdiff" / "dwi" / "sub-01_ses-phdiff_acq-HASC55AP_dwi.json"

    merge_proc = subprocess.run(
        ['bids-sidecar-merge', str(invalid_source_json), str(dest_json)])
    assert merge_proc.returncode > 0
    assert _get_json_string(dest_json) == orig_dest_json_content

def test_add_nifti_info_datalad(tmp_path):
    data_root = get_data(tmp_path)
    bod = CuBIDS(data_root / "complete", use_datalad=True, force_unlock=True)
    csv_prefix = str(tmp_path / "csvs")
    bod.get_CSVs(csv_prefix)
    summary_csv = csv_prefix + "_summary.csv"
    summary_df = pd.read_csv(summary_csv)
    l_cols = summary_df.columns.tolist()
    assert 'NumVolumes' not in l_cols
    assert 'Obliquity' not in l_cols

    # now add nifti info
    bod.add_nifti_info()
    nifti_csv_prefix = str(tmp_path / "nifti_csvs")
    bod.get_CSVs(nifti_csv_prefix)
    nifti_summary_csv = nifti_csv_prefix + "_summary.csv"
    nifti_summary_df = pd.read_csv(nifti_summary_csv)
    nifti_l_cols = nifti_summary_df.columns.tolist()
    assert 'NumVolumes' in nifti_l_cols
    assert 'Obliquity' in nifti_l_cols

def test_add_nifti_info_no_datalad(tmp_path):
    data_root = get_data(tmp_path)
    bod = CuBIDS(data_root / "complete", use_datalad=False, force_unlock=False)
    bod.add_nifti_info()
    csv_prefix = str(tmp_path / "csvs")
    bod.get_CSVs(csv_prefix)
    summary_csv = csv_prefix + "_summary.csv"
    summary_df = pd.read_csv(summary_csv)
    l_cols = summary_df.columns.tolist()
    assert 'NumVolumes' in l_cols
    assert 'Obliquity' in l_cols

#TODO: add tests that return an error for invalid merge
def test_csv_merge_no_datalad(tmp_path):
    data_root = get_data(tmp_path)
    bod = CuBIDS(data_root / "inconsistent", use_datalad=False)

    # Get an initial grouping summary and files list
    csv_prefix = str(tmp_path / "originals")
    bod.get_CSVs(csv_prefix)
    original_summary_csv = csv_prefix + "_summary.csv"
    original_files_csv = csv_prefix + "_files.csv"

    # give csv with no changes (make sure it does nothing)
    bod.apply_csv_changes(original_summary_csv,
                          original_files_csv,
                          str(tmp_path / "unmodified"))

    # these will not actually be equivalent because of the auto renames
    assert file_hash(original_summary_csv) != \
           file_hash(tmp_path / "unmodified_summary.csv")

    # Find the dwi with no FlipAngle
    summary_df = pd.read_csv(original_summary_csv)
    fa_nan_dwi_row, = np.flatnonzero(
        np.isnan(summary_df.FlipAngle) &
        summary_df.KeyGroup.str.fullmatch(
            "acquisition-HASC55AP_datatype-dwi_suffix-dwi"))
    # Find the dwi with and EchoTime ==
    complete_dwi_row, = np.flatnonzero(
        summary_df.KeyGroup.str.fullmatch(
            "acquisition-HASC55AP_datatype-dwi_suffix-dwi") &
        (summary_df.FlipAngle == 90.) &
        (summary_df.EchoTime > 0.05))
    cant_merge_echotime_dwi_row, = np.flatnonzero(
        summary_df.KeyGroup.str.fullmatch(
            "acquisition-HASC55AP_datatype-dwi_suffix-dwi") &
        (summary_df.FlipAngle == 90.) &
        (summary_df.EchoTime < 0.05))

    # Set a legal MergeInto value. This effectively fills in data
    # where there was previously as missing FlipAngle
    summary_df.loc[fa_nan_dwi_row, "MergeInto"] = summary_df.ParamGroup[
        complete_dwi_row]

    valid_csv_file = csv_prefix + "_valid_summary.csv"
    summary_df.to_csv(valid_csv_file, index=False)

    # about to apply merges!

    bod.apply_csv_changes(valid_csv_file,
                          original_files_csv,
                          str(tmp_path / "ok_modified"))

    assert not file_hash(original_summary_csv) == \
           file_hash(tmp_path / "ok_modified_summary.csv")

    # Add an illegal merge to MergeInto
    summary_df.loc[cant_merge_echotime_dwi_row, "MergeInto"] = summary_df.ParamGroup[
        complete_dwi_row]
    invalid_csv_file = csv_prefix + "_invalid_summary.csv"
    summary_df.to_csv(invalid_csv_file, index=False)

    with pytest.raises(Exception):
        bod.apply_csv_changes(invalid_csv_file,
                              str(tmp_path / "originals_files.csv"),
                              str(tmp_path / "ok_modified"))

def test_csv_merge_changes(tmp_path):
    data_root = get_data(tmp_path)
    bod = CuBIDS(data_root / "inconsistent", use_datalad=True)
    bod.datalad_save()
    assert bod.is_datalad_clean()

    # Get an initial grouping summary and files list
    csv_prefix = str(tmp_path / "originals")
    bod.get_CSVs(csv_prefix)
    original_summary_csv = csv_prefix + "_summary.csv"
    original_files_csv = csv_prefix + "_files.csv"

    # give csv with no changes (make sure it does nothing except rename)
    bod.apply_csv_changes(original_summary_csv,
                          original_files_csv,
                          str(tmp_path / "unmodified"))
    orig = pd.read_csv(original_summary_csv)
    # TEST RenameKeyGroup column got populated CORRECTLY
    for row in range(len(orig)):
        if orig.loc[row, 'ParamGroup'] != 1:
            assert str(orig.loc[row, 'RenameKeyGroup']) != 'nan'

    # TESTING RENAMES GOT APPLIED
    applied = pd.read_csv(str(tmp_path / "unmodified_summary.csv"))

    applied_f = pd.read_csv(str(tmp_path / "unmodified_files.csv"))
    odd = []
    for row in range(len(applied_f)):
        if 'VARIANT' in applied_f.loc[row, 'FilePath'] and 'VARIANT' not in applied_f.loc[row, 'KeyParamGroup']:
            odd.append((applied_f.loc[row, 'FilePath']))

    occurrences = {}
    for row in range(len(applied_f)):
        if applied_f.loc[row, 'FilePath'] in odd:
            if applied_f.loc[row, 'FilePath'] in occurrences.keys():
                occurrences[applied_f.loc[row, 'FilePath']].append(applied_f.loc[row, 'KeyParamGroup'])
            else:
                occurrences[applied_f.loc[row, 'FilePath']] = [applied_f.loc[row, 'KeyParamGroup']]

    assert len(orig) == len(applied)

    renamed = True
    new_keys = applied['KeyGroup'].tolist()
    for row in range(len(orig)):
        if orig.loc[row, 'Modality'] != 'fmap':
            if str(orig.loc[row, 'RenameKeyGroup']) != 'nan' \
                    and str(orig.loc[row, 'RenameKeyGroup']) not in new_keys:
                print(orig.loc[row, 'RenameKeyGroup'])
                renamed = False

    assert renamed == True

    # will no longer be equal because of auto rename!
    assert file_hash(original_summary_csv)!= \
           file_hash(tmp_path / "unmodified_summary.csv")

    # Find the dwi with no FlipAngle
    summary_df = pd.read_csv(original_summary_csv)
    fa_nan_dwi_row, = np.flatnonzero(
        np.isnan(summary_df.FlipAngle) &
        summary_df.KeyGroup.str.fullmatch(
            "acquisition-HASC55AP_datatype-dwi_suffix-dwi"))
    # Find the dwi with and EchoTime ==
    complete_dwi_row, = np.flatnonzero(
        summary_df.KeyGroup.str.fullmatch(
            "acquisition-HASC55AP_datatype-dwi_suffix-dwi") &
        (summary_df.FlipAngle == 90.) &
        (summary_df.EchoTime > 0.05))
    cant_merge_echotime_dwi_row, = np.flatnonzero(
        summary_df.KeyGroup.str.fullmatch(
            "acquisition-HASC55AP_datatype-dwi_suffix-dwi") &
        (summary_df.FlipAngle == 90.) &
        (summary_df.EchoTime < 0.05))

    # Set a legal MergeInto value. This effectively fills in data
    # where there was previously as missing FlipAngle
    summary_df.loc[fa_nan_dwi_row, "MergeInto"] = summary_df.ParamGroup[
        complete_dwi_row]

    valid_csv_file = csv_prefix + "_valid_summary.csv"
    summary_df.to_csv(valid_csv_file, index=False)

    # about to merge
    bod.apply_csv_changes(valid_csv_file,
                          original_files_csv,
                          str(tmp_path / "ok_modified"))

    assert not file_hash(original_summary_csv) == \
           file_hash(tmp_path / "ok_modified_summary.csv")

    # Add an illegal merge to MergeInto
    summary_df.loc[cant_merge_echotime_dwi_row, "MergeInto"] = summary_df.ParamGroup[
        complete_dwi_row]
    invalid_csv_file = csv_prefix + "_invalid_summary.csv"
    summary_df.to_csv(invalid_csv_file, index=False)

    with pytest.raises(Exception):
        bod.apply_csv_changes(invalid_csv_file,
                              str(tmp_path / "originals_files.csv"),
                              str(tmp_path / "ok_modified"))

    # Make sure MergeInto == 0 deletes the param group and all associations
    # summary_df = pd.read_csv(original_summary_csv)
    # summary_df.loc[fa_nan_dwi_row, "MergeInto"] = 0
    # delete_group = summary_df.loc[fa_nan_dwi_row, "KeyParamGroup"]

    # # files_df = pd.read_csv(original_files_csv)
    # # for row in files_df:
    # #     if files_df.iloc[row]['KeyParamGroup'] == delete_group:
    # #         filename = files_df.iloc[row]['FilePath']
    # #         file_to_rem = Path(filename)
    # #         file_to_rem.unlink()

    # delete_csv_file = csv_prefix + "_delete_summary.csv"
    # summary_df.to_csv(delete_csv_file, index=False)

    # bod.apply_csv_changes(delete_csv_file,
    #                       original_files_csv,
    #                       str(tmp_path / "ok_deleted"))

    # del_summary_csv = str(tmp_path / "ok_deleted")

    # original_summary_csv = csv_prefix + "_summary.csv"
    # original_files_csv = csv_prefix + "_files.csv"

    # assert delete_group not in tmp_path / "ok_deleted_summary.csv"

def test_merge_without_overwrite():
    meta1 = {
        'ManualCheck': 1.0,
        'RenameKeyGroup': np.nan,
        'MergeInto': 2.0,
        'KeyGroup': 'datatype-func_suffix-bold_task-rest',
        'ParamGroup': 12,
        'Counts': 2,
        'DwellTime': 2.6e-06,
        'EchoTime': 0.03,
        'EffectiveEchoSpacing': 0.000580013,
        'FieldmapKey00': 'acquisition-fMRI_datatype-fmap_direction-AP_fmap-epi_suffix-epi',
        'FieldmapKey01': 'acquisition-fMRI_datatype-fmap_direction-PA_fmap-epi_run-1_suffix-epi',
        'FieldmapKey02': 'acquisition-fMRI_datatype-fmap_direction-PA_fmap-epi_run-2_suffix-epi',
        'FieldmapKey03': np.nan,
        'FieldmapKey04': np.nan,
        'FieldmapKey05': np.nan,
        'FieldmapKey06': np.nan,
        'FieldmapKey07': np.nan,
        'FlipAngle': 31.0,
        'IntendedForKey00': np.nan,
        'IntendedForKey01': np.nan,
        'IntendedForKey02': np.nan,
        'IntendedForKey03': np.nan,
        'IntendedForKey04': np.nan,
        'IntendedForKey05': np.nan,
        'IntendedForKey06': np.nan,
        'IntendedForKey07': np.nan,
        'IntendedForKey08': np.nan,
        'IntendedForKey09': np.nan,
        'MultibandAccelerationFactor': 6.0,
        'NSliceTimes': 60,
        'ParallelReductionFactorInPlane': np.nan,
        'PartialFourier': 1.0,
        'PhaseEncodingDirection': 'j-',
        'RepetitionTime': 0.8,
        'TotalReadoutTime': 0.0481411}

    # Suppose User tries to overwrite num with NaN (allowed)
    meta_NaN = deepcopy(meta1)
    meta_NaN["FlipAngle"] = np.nan
    valid_merge = merge_without_overwrite(meta_NaN, meta1)
    assert valid_merge

    # Set a conflicting imaging param in the dest group
    meta_overwrite = deepcopy(meta1)
    meta_overwrite["FlipAngle"] = 62.0
    bad_merge = merge_without_overwrite(meta1, meta_overwrite)
    assert not bad_merge

    # Suppose "PartialFourier" is missing in the dest group
    meta_ok = deepcopy(meta1)
    meta_ok["PartialFourier"] = np.nan
    ok_merge = merge_without_overwrite(meta1, meta_ok)
    assert ok_merge
    assert ok_merge["PartialFourier"] == meta1["PartialFourier"]

    # Suppose the same, but there is a different number of slice times
    slices_bad = deepcopy(meta1)
    slices_bad["PartialFourier"] = np.nan
    slices_bad["NSliceTimes"] = meta1["NSliceTimes"] + 5
    bad_slice_merge = merge_without_overwrite(meta1, slices_bad)
    assert not bad_slice_merge


def test_keygroups(tmp_path):
    data_root = get_data(tmp_path)

    # Test the complete data
    complete_bod = CuBIDS(data_root / "complete")
    complete_misfit_fmaps = complete_bod._cache_fieldmaps()
    # There should be no unpaired fieldmaps
    assert len(complete_misfit_fmaps) == 0

    # Test that the correct key groups are found
    key_groups = complete_bod.get_key_groups()
    assert key_groups == COMPLETE_KEY_GROUPS

    # Test the incomplete
    ibod = CuBIDS(data_root / "inconsistent")
    inc_misfit_fmaps = ibod._cache_fieldmaps()
    assert len(inc_misfit_fmaps) == 1

    # There will still be the same number of key groups
    ikey_groups = ibod.get_key_groups()
    assert ikey_groups == COMPLETE_KEY_GROUPS


def test_csv_creation(tmp_path):
    """Test the Key Group and Parameter Group creation on sample data.
    """
    data_root = get_data(tmp_path)

    # Test the complete data
    complete_bod = CuBIDS(data_root / "complete")
    complete_misfit_fmaps = complete_bod._cache_fieldmaps()
    # There should be no unpaired fieldmaps
    assert len(complete_misfit_fmaps) == 0

    # Test that the correct key groups are found
    key_groups = complete_bod.get_key_groups()
    assert key_groups == COMPLETE_KEY_GROUPS

    # Get the CSVs from the complete data
    cfiles_df, csummary_df = \
        complete_bod.get_param_groups_dataframes()

    # Make sure we got all 21 of the files
    assert cfiles_df.shape[0] == 21

    # This data should have the same number of param
    # groups as key groups
    assert csummary_df.shape[0] == len(COMPLETE_KEY_GROUPS)

    # check IntendedForXX and FieldmapKeyXX are boolean now
    bool_IF = False
    bool_FMAP = False
    for row in range(len(csummary_df)):
        if str(csummary_df.loc[row, "UsedAsFieldmap"]) == "True":
            bool_IF = True
        if str(csummary_df.loc[row, "HasFieldmap"]) == "True":
            bool_FMAP = True
    assert bool_IF == True
    assert bool_FMAP == True

    # Test the incomplete
    ibod = CuBIDS(data_root / "inconsistent")
    inc_misfit_fmaps = ibod._cache_fieldmaps()
    assert len(inc_misfit_fmaps) == 1

    # There will still be the same number of key groups
    ikey_groups = ibod.get_key_groups()
    assert ikey_groups == COMPLETE_KEY_GROUPS

    # Get the CSVs from the inconsistent data
    ifiles_df, isummary_df = \
        ibod.get_param_groups_dataframes()

    # There are still 21 files
    assert ifiles_df.shape[0] == 21

    # But now there are more parameter groups
    assert isummary_df.shape[0] == 12

    # check that summary csv param group nums are in the right order
    # and check that param groups are sorted by count vals
    for i, (index, row) in enumerate(isummary_df.iterrows()):
        if i == len(isummary_df) -1:
            break
        # if key groups in rows i and i+1 are the same
        if isummary_df.iloc[i]['KeyGroup'] == \
            isummary_df.iloc[i+1]['KeyGroup']:
            # param group i = param group i+1
            assert isummary_df.iloc[i]['ParamGroup'] == \
                isummary_df.iloc[i+1]['ParamGroup'] - 1
            # and count i < count i + 1
            assert isummary_df.iloc[i]['Counts'] >= \
                isummary_df.iloc[i+1]['Counts']

    # check that files csv param group nums are in the right order
    for i, (index, row) in enumerate(ifiles_df.iterrows()):
        if i == len(ifiles_df) -1:
            break
        # if key groups in rows i and i+1 are the same
        if ifiles_df.iloc[i]['KeyGroup'] == \
            ifiles_df.iloc[i+1]['KeyGroup']:
            # param group i = param group i+1
            assert ifiles_df.iloc[i]['ParamGroup'] <= \
                ifiles_df.iloc[i+1]['ParamGroup']


def test_apply_csv_changes(tmp_path):
    # set up like narrative of user using this
    # similar to test csv creation
    # open the csv, rename a key group
    # save csv
    # call change key groups
    # give csv with no changes (make sure it does nothing)
    # make sure files you wanted to rename exist in the bids dir

    data_root = get_data(tmp_path)
    bids_dir = str(data_root / "complete")
    for scan in Path(bids_dir).rglob('sub-*/*/*/*.nii.gz'):

        # add extension files
        _add_ext_files(str(scan))

    # path_to_img = str(data_root / "complete/sub-01/ses-phdiff/fmap/sub-01_ses-phdiff_acq-v4_magnitude1.nii.gz")
    # _add_ext_files(path_to_img)

    has_events = False
    has_physio = False

    # check if events and physio files
    if Path(data_root /
            "complete/sub-01/ses-phdiff/fmap/sub-01_ses-phdiff_acq-v4_events.tsv").exists():
        has_events = True
    if Path(data_root /
            "complete/sub-01/ses-phdiff/fmap/sub-01_ses-phdiff_acq-v4_physio.tsv.gz").exists():
        has_physio = True


    complete_cubids = CuBIDS(data_root / "complete", use_datalad=True)
    complete_cubids.datalad_save()

    complete_cubids.get_CSVs(str(tmp_path / "originals"))

    # give csv with no changes (make sure it does nothing)
    complete_cubids.apply_csv_changes(str(tmp_path / "originals_summary.csv"),
                                    str(tmp_path / "originals_files.csv"),
                                    str(tmp_path / "modified1"))

    og_path = tmp_path / "originals_summary.csv"
    with og_path.open("r") as f:
        og_content = "".join(f.readlines())

    mod1_path = tmp_path / "modified1_summary.csv"
    with mod1_path.open("r") as f:
        mod1_content = "".join(f.readlines())

    assert og_content == mod1_content

    # edit the csv, add a RenameKeyGroup

    # _edit_csv(str(tmp_path / "originals_summary.csv"))
    complete_cubids.apply_csv_changes(str(tmp_path / "originals_summary.csv"),
                                    str(tmp_path / "originals_files.csv"),
                                    str(tmp_path / "modified2"))

    # check files df to make sure extension files also got renmaed
    mod_files = tmp_path / "modified2_files.csv"
    # ensure fmap didn't get renamed
    # assert Path(data_root /
    #     "complete/sub-01/ses-phdiff/fmap/sub-01_ses-phdiff_acq-v5_magnitude1.json").exists() == False
    assert Path(data_root /
        "complete/sub-01/ses-phdiff/fmap/sub-01_ses-phdiff_acq-v4_magnitude1.json").exists() == True

    # check that old names are gone!
    # assert Path(data_root /
    #     "complete/sub-01/ses-phdiff/fmap/sub-01_ses-phdiff_acq-v5_physio.tsv.gz").exists() == True
    assert Path(data_root /
        "complete/sub-01/ses-phdiff/fmap/sub-01_ses-phdiff_acq-v4_physio.tsv.gz").exists() == True

    mod2_path = tmp_path / "modified2_summary.csv"
    with mod2_path.open("r") as f:
        mod2_content = "".join(f.readlines())

    assert og_content == mod2_content

    # check that MergeInto = 0 deletes scan and associations
    deleted_keyparam = _add_deletion(mod2_path)
    assert deleted_keyparam in mod2_content

    # check to delete keyparam  exist
    mod2_files = tmp_path / "modified2_files.csv"
    with mod2_files.open("r") as f:
        mod2_f_content = "".join(f.readlines())
    assert deleted_keyparam in mod2_f_content

    # check scans and associations to be deleted are currently in the bids dir
    mod2_summary_df = pd.read_csv(mod2_path)
    mod2_files_df = pd.read_csv(str(tmp_path / "modified2_files.csv"))
    deleted_f = []

    for row in range(len(mod2_files_df)):
        if mod2_files_df.loc[row, 'KeyParamGroup'] == deleted_keyparam:
            deleted_f.append(mod2_files_df.loc[row, 'FilePath'])

    for f in deleted_f:
        assert Path(f).exists() == True
        assert Path(f.replace('nii.gz', 'json')).exists() == True

    # apply deletion
    complete_cubids.apply_csv_changes(mod2_path,
                                    str(tmp_path / "modified2_files.csv"),
                                    str(tmp_path / "deleted"))

    # make sure deleted_keyparam gone from files_csv
    deleted = tmp_path / "deleted_summary.csv"
    with deleted.open("r") as f:
        deleted_content = "".join(f.readlines())
    assert deleted_keyparam not in deleted_content

    # make sure deleted_keyparam gone from summary csv
    deleted_files = tmp_path / "deleted_files.csv"
    with deleted_files.open("r") as f:
        deleted_f_content = "".join(f.readlines())
    assert deleted_keyparam not in deleted_f_content

    # make sure deleted files are gone
    for f in deleted_f:
        assert Path(f).exists() == False
        assert Path(f.replace('nii.gz', 'json')).exists() == False


def test_session_apply(tmp_path):
    # set up like narrative of user using this
    # similar to test csv creation
    # open the csv, rename a key group
    # save csv
    # call change key groups
    # give csv with no changes (make sure it does nothing)
    # make sure files you wanted to rename exist in the bids dir

    data_root = get_data(tmp_path)
    bids_dir = str(data_root / "inconsistent")

    ses_cubids = CuBIDS(data_root / "inconsistent", acq_group_level='session', use_datalad=True)

    ses_cubids.get_CSVs(str(tmp_path / "originals"))

    # give csv and make sure 'session' is in summary both pre and post apply
    ses_cubids.apply_csv_changes(str(tmp_path / "originals_summary.csv"),
                                    str(tmp_path / "originals_files.csv"),
                                    str(tmp_path / "modified1"))

    og_path = tmp_path / "originals_summary.csv"
    with og_path.open("r") as f:
        og_content = "".join(f.readlines())

    mod1_path = tmp_path / "modified1_summary.csv"
    with mod1_path.open("r") as f:
        mod1_content = "".join(f.readlines())

    assert 'session-' in og_content
    assert 'session-' in mod1_content



def _add_deletion(summary_csv):
    df = pd.read_csv(summary_csv)
    df.loc[3, 'MergeInto'] = 0
    df.to_csv(summary_csv, index=False)
    return df.loc[3, 'KeyParamGroup']


# def _edit_csv(summary_csv):
#     df = pd.read_csv(summary_csv)
#     df['RenameKeyGroup'] = df['RenameKeyGroup'].apply(str)
#     df['KeyGroup'] = df['KeyGroup'].apply(str)
#     for row in range(len(df)):
#         if df.loc[row, 'KeyGroup'] == \
#             "acquisition-v4_datatype-fmap_fmap-magnitude1_suffix-magnitude1":
#             df.at[row, 'RenameKeyGroup'] = \
#                 "acquisition-v5_datatype-fmap_fmap-magnitude1_suffix-magnitude1"
#     df.to_csv(summary_csv)

def _add_ext_files(img_path):
    # add and save extension files in
    dwi_exts = ['.bval', '.bvec']

    # everyone gets a physio file
    no_suffix = img_path.rpartition('_')[0]
    physio_file = no_suffix + '_physio' + '.tsv.gz'
    # save ext file in img_path's parent dir
    Path(physio_file).touch()

    if '/dwi/' in img_path:
        # add bval and bvec
        for ext in dwi_exts:
            dwi_ext_file = img_path.replace(".nii.gz", "").replace(".nii", "") + ext
            Path(dwi_ext_file).touch()
    if 'bold' in img_path:
        no_suffix = img_path.rpartition('_')[0]
        bold_ext_file = no_suffix + '_events' + '.tsv'
        Path(bold_ext_file).touch()


def _edit_a_json(json_file):
    """Open a json file, write somthing to it and save it to the same name."""
    with open(json_file, "r") as metadatar:
        metadata = json.load(metadatar)

    metadata["THIS_IS_A_TEST"] = True
    with open(json_file, "w") as metadataw:
        json.dump(metadata, metadataw)


def _edit_a_nifti(nifti_file):
    img = nb.load(nifti_file)
    new_img = nb.Nifti1Image(np.random.rand(*img.shape),
                             affine=img.affine,
                             header=img.header)
    new_img.to_filename(nifti_file)


def file_hash(file_name):
    with open(str(file_name), 'rb') as fcheck:
        data = fcheck.read()
    return hashlib.md5(data).hexdigest()


def _get_json_string(json_path):
    with json_path.open("r") as f:
        content = "".join(f.readlines())
    return content


def test_remove_fields(tmp_path):
    """Test that we metadata fields are detected and removed."""
    data_root = get_data(tmp_path)
    bod = CuBIDS(data_root, use_datalad=False)

    # Get the metadata fields
    metadata_fields = bod.get_all_metadata_fields()
    assert metadata_fields

    # Simulate some fields we might want to remove
    fields_to_remove = ["DeviceSerialNumber", "AcquisitionTime",
                        "InstitutionAddress", "InstitutionName",
                        "StationName", "NotARealField"]

    bod.remove_metadata_fields(fields_to_remove)
    new_fields = bod.get_all_metadata_fields()
    assert not set(new_fields).intersection(fields_to_remove)


def test_datalad_integration(tmp_path):
    """Test that datalad works for basic file modification operations.
    """
    data_root = get_data(tmp_path)

    # Test that an uninitialized CuBIDS raises exceptions
    uninit_cubids = CuBIDS(data_root / "complete", use_datalad=False)

    # Ensure an exception is raised if trying to use datalad without
    # initializing
    with pytest.raises(Exception):
        uninit_cubids.is_datalad_clean()

    # initialize the datalad repository and try again
    uninit_cubids.init_datalad()
    uninit_cubids.datalad_save('Test save')
    assert uninit_cubids.is_datalad_clean()

    # Now, the datalad repository is initialized and saved.
    # Make sure if we make a new CuBIDS object it recognizes that
    # the datalad status is OK
    complete_bod = CuBIDS(data_root / "complete", use_datalad=True)

    assert complete_bod.datalad_ready
    assert complete_bod.is_datalad_clean()

    # Test clean and revert functionality
    test_file = data_root / "complete" / "sub-03" / "ses-phdiff" \
        / "func" / "sub-03_ses-phdiff_task-rest_bold.json"
    test_binary = data_root / "complete" / "sub-03" / "ses-phdiff" \
        / "func" / "sub-03_ses-phdiff_task-rest_bold.nii.gz"

    # Try editing a locked file - it should fail
    with pytest.raises(Exception):
        _edit_a_nifti(test_binary)

    # Unlock the files so we can access their content
    complete_bod.datalad_handle.unlock(test_binary)
    complete_bod.datalad_handle.unlock(test_file)

    # Get the contents of the original files
    original_content = _get_json_string(test_file)
    original_binary_content = file_hash(test_binary)

    # Edit the files
    _edit_a_nifti(test_binary)
    _edit_a_json(test_file)

    # Get the edited content
    edited_content = _get_json_string(test_file)
    edited_binary_content = file_hash(test_binary)

    # Check that the file content has changed
    assert not original_content == edited_content
    assert not original_binary_content == edited_binary_content

    # Check that datalad knows something has changed
    assert not uninit_cubids.is_datalad_clean()
    assert not complete_bod.is_datalad_clean()

    # Attempt to undo a change before checking in changes
    with pytest.raises(Exception):
        uninit_cubids.datalad_undo_last_commit()

    # Perform a save
    uninit_cubids.datalad_save(message="TEST SAVE!")
    assert uninit_cubids.is_datalad_clean()

    # Now undo the most recent save
    uninit_cubids.datalad_undo_last_commit()

    # Unlock the restored files so we can access their content
    complete_bod.datalad_handle.unlock(test_binary)
    complete_bod.datalad_handle.unlock(test_file)

    # Get the contents of the original files
    restored_content = _get_json_string(test_file)
    restored_binary_content = file_hash(test_binary)

    # Check that the file content has returned to its original state
    assert original_content == restored_content
    assert original_binary_content == restored_binary_content


def _remove_a_json(json_file):

    os.remove(json_file)


def test_validator(tmp_path):

    data_root = get_data(tmp_path)

    # test the validator in valid dataset
    call = build_validator_call(str(data_root) + "/complete")
    ret = run_validator(call)

    assert ret.returncode == 0
    parsed = parse_validator_output(ret.stdout.decode('UTF-8'))


    # change this assert
    # assert parsed.shape[1] < 1


    # bungle some data and test

    # get data
    test_file = data_root / "complete" / "sub-03" / "ses-phdiff" \
        / "func" / "sub-03_ses-phdiff_task-rest_bold.json"
    test_binary = data_root / "complete" / "sub-03" / "ses-phdiff" \
        / "func" / "sub-03_ses-phdiff_task-rest_bold.nii.gz"

    # Edit the files
    _edit_a_nifti(test_binary)
    _remove_a_json(test_file)

    call = build_validator_call(str(data_root) + "/complete")
    ret = run_validator(call)

    assert ret.returncode == 1

    parsed = parse_validator_output(ret.stdout.decode('UTF-8'))

    assert type(parsed) == pd.core.frame.DataFrame


def test_docker():
    """Verify that docker is installed and the user has permission to
    run docker images.
    Returns
    -------
    -1  Docker can't be found
     0  Docker found, but user can't connect to daemon
     1  Test run OK
     """
    try:

        return_status = 1
        ret = subprocess.run(['docker', 'version'], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    except OSError as e:
        from errno import ENOENT
        if e.errno == ENOENT:
            print("Cannot find Docker engine!")
            return_status = 0
        raise e
    if ret.stderr.startswith(b"Cannot connect to the Docker daemon."):
        print("Cannot connect to Docker daemon!")
        return_status = 0
    assert return_status


# def test_image(image='pennlinc/bond:latest'):
#     """Check whether image is present on local system"""
#     ret = subprocess.run(['docker', 'images', '-q', image],
#                          stdout=subprocess.PIPE)

#     return_status = ret.stdout.decode('UTF-8')
#     assert return_status



"""Main module."""
import json
from collections import defaultdict
import numpy as np
import pandas as pd
from copy import deepcopy
from math import nan, isnan
from .constants import IMAGING_PARAMS
DIRECT_IMAGING_PARAMS = IMAGING_PARAMS - set(["NSliceTimes"])


def check_merging_operations(action_csv, raise_on_error=False):
    """Checks that the merges in an action csv are possible.

    To be mergable the
    """
    actions = pd.read_csv(action_csv)
    ok_merges = []
    deletions = []
    overwrite_merges = []
    sdc_incompatible = []

    sdc_cols = set([col for col in actions.columns if
                    col.startswith("IntendedForKey") or
                    col.startswith("FieldmapKey")])

    def _check_sdc_cols(meta1, meta2):
        return {key: meta1[key] for key in sdc_cols} == \
               {key: meta2[key] for key in sdc_cols}

    needs_merge = actions[np.isfinite(actions['MergeInto'])]
    for _, row_needs_merge in needs_merge.iterrows():
        source_param_key = tuple(row_needs_merge[["MergeInto", "KeyGroup"]])
        dest_param_key = tuple(row_needs_merge[["ParamGroup", "KeyGroup"]])
        dest_metadata = row_needs_merge.to_dict()
        source_row = actions.loc[
            (actions[["ParamGroup", "KeyGroup"]] == source_param_key).all(1)]

        if source_param_key[0] == 0:
            print("going to delete ", dest_param_key)
            deletions.append(dest_param_key)
            continue
        if not source_row.shape[0] == 1:
            raise Exception("Could not identify a unique source group")
        source_metadata = source_row.iloc[0].to_dict()
        merge_id = (source_param_key, dest_param_key)
        # Check for compatible fieldmaps
        if not _check_sdc_cols(source_metadata, dest_metadata):
            sdc_incompatible.append(merge_id)
            continue

        if not merge_without_overwrite(source_metadata, dest_metadata,
                                       raise_on_error=raise_on_error):
            overwrite_merges.append(merge_id)
            continue
        # add to the list of ok merges if there are no conflicts
        ok_merges.append(merge_id)

    error_message = "\n\nProblems were found in the requested merge.\n" \
                    "===========================================\n\n"
    if sdc_incompatible:
        error_message += "Some merges are incompatible due to differing " \
                         "distortion correction strategies. Check that " \
                         "fieldmaps exist and have the correct " \
                         "\"IntendedFor\" in their sidecars. These merges " \
                         "could not be completed:\n"
        error_message += print_merges(sdc_incompatible) + "\n\n"

    if overwrite_merges:
        error_message += "Some merges are incompatible because the metadata " \
                         "in the destination json conflicts with the values " \
                         "in the source json. Merging should only be used " \
                         "to fill in missing metadata. The following " \
                         "merges could not be completed:\n\n"
        error_message += print_merges(overwrite_merges)

    if overwrite_merges or sdc_incompatible:
        if raise_on_error:
            raise Exception(error_message)
        print(error_message)
    return ok_merges, deletions


def merge_without_overwrite(source_meta, dest_meta_orig, raise_on_error=False):
    """Performs a safe metadata copy.

    Here, "safe" means that no non-NaN values in `dest_meta` are
    overwritten by the merge. If any overwrites occur an empty
    dictionary is returned.
    """
    # copy the original json params
    dest_meta = deepcopy(dest_meta_orig)

    if not source_meta.get("NSliceTimes") == dest_meta.get("NSliceTimes"):
        if raise_on_error:
            raise Exception("Value for NSliceTimes is %d in destination "
                            "but %d in source"
                            % (source_meta.get("NSliceTimes"),
                               source_meta.get("NSliceTimes")))
        return {}
    for parameter in DIRECT_IMAGING_PARAMS:
        source_value = source_meta.get(parameter, nan)
        dest_value = dest_meta.get(parameter, nan)

        # cannot merge num --> num
        # exception should only be raised
        # IF someone tries to replace a num (dest)
        # with a num (src)
        if not is_nan(source_value):
            # need to figure out if we can merge
            if not is_nan(dest_value) and source_value != dest_value:
                if raise_on_error:
                    raise Exception("Value for %s is %s in destination "
                                    "but %s in source"
                                    % (parameter, str(dest_value),
                                       str(source_value)))
                return {}
        dest_meta[parameter] = source_value
    return dest_meta


def is_nan(val):
    '''Returns True if val is nan'''
    if not isinstance(val, float):
        return False
    return isnan(val)


def print_merges(merge_list):
    """Print formatted text of merges"""
    return "\n\t" + "\n\t".join(
        ["%s \n\t\t-> %s" % ("%s:%d" % src_id[::-1],
         "%s:%d" % dest_id[::-1]) for
         src_id, dest_id in merge_list])


def merge_json_into_json(from_file, to_file,
                         raise_on_error=False):
    print("Merging imaging metadata from %s to %s"
          % (from_file, to_file))
    with open(from_file, "r") as fromf:
        source_metadata = json.load(fromf)

    with open(to_file, "r") as tof:
        dest_metadata = json.load(tof)
    orig_dest_metadata = deepcopy(dest_metadata)

    merged_metadata = merge_without_overwrite(
        source_metadata, dest_metadata, raise_on_error=raise_on_error)

    if not merged_metadata:
        return 255

    # Only write if the data has changed
    if not merged_metadata == orig_dest_metadata:
        print("OVERWRITING", to_file)
        with open(to_file, "w") as tofw:
            json.dump(merged_metadata, tofw, indent=4)

    return 0


def group_by_acquisition_sets(files_csv, output_prefix, acq_group_level):
    '''Finds unique sets of Key/Param groups across subjects.
    '''
    from bids.layout import parse_file_entities
    from bids import config
    config.set_option('extension_initial_dot', True)

    files_df = pd.read_csv(files_csv)
    acq_groups = defaultdict(list)
    for _, row in files_df.iterrows():
        file_entities = parse_file_entities(row.FilePath)

        if acq_group_level == 'subject':
            acq_id = (file_entities.get("subject"),
                      file_entities.get("session"))
            acq_groups[acq_id].append((row.KeyGroup, row.ParamGroup))
        else:
            acq_id = (file_entities.get("subject"), None)
            acq_groups[acq_id].append((row.KeyGroup, row.ParamGroup,
                                       file_entities.get("session")))

    # Map the contents to a list of subjects/sessions
    contents_to_subjects = defaultdict(list)
    for key, value in acq_groups.items():
        contents_to_subjects[tuple(sorted(value))].append(key)

    # Sort them based on how many have that group
    content_ids = []
    content_id_counts = []
    for key, value in contents_to_subjects.items():
        content_ids.append(key)
        content_id_counts.append(len(value))

    descending_order = np.argsort(content_id_counts)[::-1]

    # Create a dataframe with the subject, session, groupnum
    grouped_sub_sess = []
    acq_group_info = []
    for groupnum, content_id_row in enumerate(descending_order, start=1):
        content_id = content_ids[content_id_row]
        acq_group_info.append(
            (groupnum, content_id_counts[content_id_row]) + content_id)
        for subject, session in contents_to_subjects[content_id]:
            grouped_sub_sess.append(
                {"subject": 'sub-' + subject,
                 "session": session,
                 "AcqGroup": groupnum})

    # Write the mapping of subject/session to
    acq_group_df = pd.DataFrame(grouped_sub_sess)
    acq_group_df.to_csv(output_prefix + "_AcqGrouping.csv", index=False)

    # Write the summary of acq groups to a text file
    with open(output_prefix + "_AcqGroupInfo.txt", "w") as infotxt:
        infotxt.write(
            "\n".join([" ".join(map(str, line)) for line in acq_group_info]))



import subprocess
import json
import logging
import os
import glob
import pathlib
import pandas as pd

logger = logging.getLogger('cubids-cli')


def build_validator_call(path, ignore_headers=False, ignore_subject=True):
    """Build a subprocess command to the bids validator"""

    # build docker call
    command = ['bids-validator', '--verbose', '--json']

    if ignore_headers:
        command.append('--ignoreNiftiHeaders')
    if ignore_subject:
        command.append('--ignoreSubjectConsistency')

    command.append(path)

    return command


def build_subject_paths(bids_dir):
    """Build a list of BIDS dirs with 1 subject each"""

    bids_dir = str(bids_dir)
    if not bids_dir.endswith("/"):
        bids_dir += "/"

    root_files = [x for x in glob.glob(bids_dir + "*") if os.path.isfile(x)]

    bids_dir += "sub-*/"

    subjects = glob.glob(bids_dir)

    if len(subjects) < 1:

        raise ValueError("Couldn't find any subjects "
                         "in the specified directory:\n" +
                         bids_dir)

    subjects_dict = {}

    for sub in subjects:
        purepath = pathlib.PurePath(sub)
        sub_label = purepath.name

        files = [x for x in glob.glob(sub + '**', recursive=True)
                 if os.path.isfile(x)]
        files.extend(root_files)
        subjects_dict[sub_label] = files

    return subjects_dict


def run_validator(call, verbose=True):
    """Run the validator with subprocess"""
    # if verbose:
    #     logger.info("Running the validator with call:")
    #     logger.info('\"' + ' '.join(call) + '\"')

    ret = subprocess.run(call, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    return(ret)


def parse_validator_output(output):
    """Parse the JSON output of the BIDS validator into a pandas dataframe
    Parameters:
    -----------
        - path : string
            Path to JSON file of BIDS validator output
    Returns
    -----------
        - Pandas DataFrame
    """

    def get_nested(dct, *keys):
        for key in keys:
            try:
                dct = dct[key]
            except (KeyError, TypeError):
                return None
        return dct

    data = json.loads(output)

    issues = data['issues']

    def parse_issue(issue_dict):

        return_dict = {}
        return_dict['files'] = [
            get_nested(x, 'file', 'relativePath')
            for x in issue_dict.get('files', '')
            ]
        return_dict['type'] = issue_dict.get('key' '')
        return_dict['severity'] = issue_dict.get('severity', '')
        return_dict['description'] = issue_dict.get('reason', '')
        return_dict['code'] = issue_dict.get('code', '')
        return_dict['url'] = issue_dict.get('helpUrl', '')

        return(return_dict)

    df = pd.DataFrame()

    for warn in issues['warnings']:

        parsed = parse_issue(warn)
        parsed = pd.DataFrame(parsed)
        df = pd.concat([df, parsed], ignore_index=True)

    for err in issues['errors']:

        parsed = parse_issue(err)
        parsed = pd.DataFrame(parsed)
        df = pd.concat([df, parsed], ignore_index=True)

    return df



"""
Functions for configuring CuBIDS
"""

from pkg_resources import resource_filename as pkgrf
import yaml
from pathlib import Path


def load_config(config_file):
    """Loads a YAML file containing a configuration for param groups.
    """

    if config_file is None:
        config_file = Path(pkgrf("cubids", "data/config.yml"))

    with config_file.open() as f:
        config = yaml.safe_load(f)

    return config



"""Main module."""
import warnings
from collections import defaultdict
import subprocess
import bids
import bids.layout
import json
import csv
import os
from pathlib import Path
from bids.layout import parse_file_entities
from bids.utils import listify
import numpy as np
import pandas as pd
import nibabel as nb
import datalad.api as dlapi
from shutil import copytree, copyfile
from sklearn.cluster import AgglomerativeClustering
from tqdm import tqdm
from .constants import ID_VARS, NON_KEY_ENTITIES
from .config import load_config
from .metadata_merge import (
    check_merging_operations, group_by_acquisition_sets)
warnings.simplefilter(action='ignore', category=FutureWarning)
bids.config.set_option('extension_initial_dot', True)


class CuBIDS(object):

    def __init__(self, data_root, use_datalad=False, acq_group_level='subject',
                 grouping_config=None, force_unlock=False):

        self.path = os.path.abspath(data_root)
        self._layout = None
        self.keys_files = {}
        self.fieldmaps_cached = False
        self.datalad_ready = False
        self.datalad_handle = None
        self.old_filenames = []  # files whose key groups changed
        self.new_filenames = []  # new filenames for files to change
        self.IF_rename_paths = []  # fmap jsons with rename intended fors
        self.grouping_config = load_config(grouping_config)
        self.acq_group_level = acq_group_level
        self.scans_txt = None  # txt file of scans to purge (for purge only)
        self.force_unlock = force_unlock  # force unlock for add-nifti-info

        self.use_datalad = use_datalad  # True if flag set, False if flag unset
        if self.use_datalad:
            self.init_datalad()

        if self.acq_group_level == 'session':
            NON_KEY_ENTITIES.remove("session")

    @property
    def layout(self):
        if self._layout is None:
            self.reset_bids_layout()
        return self._layout

    def reset_bids_layout(self, validate=False):
        self._layout = bids.BIDSLayout(self.path, validate=validate)

    def init_datalad(self):
        """Initializes a datalad Dataset at self.path.

        Parameters:
        -----------

            save: bool
                Run datalad save to add any untracked files
            message: str or None
                Message to add to
        """
        self.datalad_ready = True

        self.datalad_handle = dlapi.Dataset(self.path)
        if not self.datalad_handle.is_installed():
            self.datalad_handle = dlapi.create(self.path,
                                               cfg_proc='text2git',
                                               force=True,
                                               annex=True)

    def datalad_save(self, message=None):
        """Performs a DataLad Save operation on the BIDS tree.

        Additionally a check for an active datalad handle and that the
        status of all objects after the save is "ok".

        Parameters:
        -----------
            message : str or None
                Commit message to use with datalad save
        """

        if not self.datalad_ready:
            raise Exception(
                "DataLad has not been initialized. use datalad_init()")
        statuses = self.datalad_handle.save(message=message or "CuBIDS Save")
        saved_status = set([status['status'] for status in statuses])
        if not saved_status == set(["ok"]):
            raise Exception("Failed to save in DataLad")

    def is_datalad_clean(self):
        """If True, no changes are detected in the datalad dataset."""
        if not self.datalad_ready:
            raise Exception(
                "Datalad not initialized, can't determine status")
        statuses = set([status['state'] for status in
                        self.datalad_handle.status()])
        return statuses == set(["clean"])

    def datalad_undo_last_commit(self):
        """Revert the most recent commit, remove it from history.

        Uses git reset --hard to revert to the previous commit.
        """
        if not self.is_datalad_clean():
            raise Exception("Untracked changes present. "
                            "Run clear_untracked_changes first")
        reset_proc = subprocess.run(
            ["git", "reset", "--hard", "HEAD~1"], cwd=self.path)
        reset_proc.check_returncode()

    def add_nifti_info(self, raise_on_error=True):
        """Adds info from nifti files to json sidecars."""
        # check if force_unlock is set
        if self.force_unlock:
            # CHANGE TO SUBPROCESS.CALL IF NOT BLOCKING
            subprocess.run(["datalad", "unlock"], cwd=self.path)

        # loop through all niftis in the bids dir
        for path in Path(self.path).rglob("sub-*/**/*.*"):
            # ignore all dot directories
            if '/.' in str(path):
                continue
            if str(path).endswith(".nii") or str(path).endswith(".nii.gz"):
                try:
                    img = nb.load(str(path))
                except Exception:
                    print("Empty Nifti File: ", str(path))
                    continue
                # get important info from niftis
                obliquity = np.any(nb.affines.obliquity(img.affine)
                                   > 1e-4)
                voxel_sizes = img.header.get_zooms()
                matrix_dims = img.shape
                # add nifti info to corresponding sidecars​
                sidecar = img_to_new_ext(str(path), '.json')
                if Path(sidecar).exists():

                    with open(sidecar) as f:
                        data = json.load(f)

                    if "Obliquity" not in data.keys():
                        data["Obliquity"] = str(obliquity)
                    if "VoxelSizeDim1" not in data.keys():
                        data["VoxelSizeDim1"] = float(voxel_sizes[0])
                    if "VoxelSizeDim2" not in data.keys():
                        data["VoxelSizeDim2"] = float(voxel_sizes[1])
                    if "VoxelSizeDim3" not in data.keys():
                        data["VoxelSizeDim3"] = float(voxel_sizes[2])
                    if "Dim1Size" not in data.keys():
                        data["Dim1Size"] = matrix_dims[0]
                    if "Dim2Size" not in data.keys():
                        data["Dim2Size"] = matrix_dims[1]
                    if "Dim3Size" not in data.keys():
                        data["Dim3Size"] = matrix_dims[2]
                    if "NumVolumes" not in data.keys():
                        if img.ndim == 4:
                            data["NumVolumes"] = matrix_dims[3]
                        elif img.ndim == 3:
                            data["NumVolumes"] = 1.0
                    with open(sidecar, 'w') as file:
                        json.dump(data, file, indent=4)

        if self.use_datalad:
            self.datalad_save(message="Added nifti info to sidecars")
            self.reset_bids_layout()

    def apply_csv_changes(self, summary_csv, files_csv, new_prefix,
                          raise_on_error=True):
        """Applies changes documented in the edited _summary csv
        and generates the new csv files.

        This function looks at the RenameKeyGroup and MergeInto
        columns and modifies the bids datset according to the
        specified changs.

        Parameters:
        -----------
            orig_prefix : str
                Path prefix and file stem for the original
                _summary and _files csvs.
                For example, if orig_prefix is
                '/cbica/projects/HBN/old_CSVs' then the paths to
                the summary and files csvs will be
                '/cbica/projects/HBN/old_CSVs_summary.csv' and
                '/cbica/projects/HBN/old_CSVs_files.csv' respectively.
            new_prefix : str
                Path prefix and file stem for the new summary and
                files csvs.
        """
        # reset lists of old and new filenames
        self.old_filenames = []
        self.new_filenames = []

        files_df = pd.read_csv(files_csv)
        summary_df = pd.read_csv(summary_csv)

        # Check that the MergeInto column only contains valid merges
        ok_merges, deletions = check_merging_operations(
            summary_csv, raise_on_error=raise_on_error)

        merge_commands = []
        for source_id, dest_id in ok_merges:
            dest_files = files_df.loc[
                (files_df[["ParamGroup", "KeyGroup"]] == dest_id).all(1)]
            source_files = files_df.loc[
                (files_df[["ParamGroup", "KeyGroup"]] == source_id).all(1)]

            # Get a source json file
            source_json = img_to_new_ext(source_files.iloc[0].FilePath,
                                         '.json')
            for dest_nii in dest_files.FilePath:
                dest_json = img_to_new_ext(dest_nii, '.json')
                if Path(dest_json).exists() and Path(source_json).exists():
                    merge_commands.append(
                        'bids-sidecar-merge %s %s'
                        % (source_json, dest_json))

        # Get the delete commands
        # delete_commands = []
        to_remove = []
        for rm_id in deletions:
            files_to_rm = files_df.loc[
                (files_df[["ParamGroup", "KeyGroup"]] == rm_id).all(1)]
            for rm_me in files_to_rm.FilePath:
                if Path(rm_me).exists():
                    to_remove.append(rm_me)
                    # delete_commands.append("rm " + rm_me)

        # call purge associations on list of files to remove
        self._purge_associations(to_remove)

        # Now do the file renaming
        change_keys_df = summary_df[summary_df.RenameKeyGroup.notnull()]
        move_ops = []
        # return if nothing to change
        if len(change_keys_df) > 0:

            key_groups = {}

            for i in range(len(change_keys_df)):
                new_key = change_keys_df.iloc[i]['RenameKeyGroup']
                old_key_param = change_keys_df.iloc[i]['KeyParamGroup']

                # add to dictionary
                key_groups[old_key_param] = new_key

            # orig key/param tuples that will have new key group
            to_change = list(key_groups.keys())

            for row in range(len(files_df)):
                file_path = files_df.loc[row, 'FilePath']
                if Path(file_path).exists() and '/fmap/' not in file_path:

                    key_param_group = files_df.loc[row, 'KeyParamGroup']

                    if key_param_group in to_change:

                        orig_key_param = files_df.loc[row, 'KeyParamGroup']

                        new_key = key_groups[orig_key_param]

                        new_entities = _key_group_to_entities(new_key)

                        # generate new filenames according to new key group
                        self.change_filename(file_path, new_entities)

            # create string of mv command ; mv command for dlapi.run
            for from_file, to_file in zip(self.old_filenames,
                                          self.new_filenames):

                if Path(from_file).exists():
                    # if using datalad, we want to git mv instead of mv
                    if self.use_datalad:
                        move_ops.append('git mv %s %s' % (from_file, to_file))
                    else:
                        move_ops.append('mv %s %s' % (from_file, to_file))
        full_cmd = "\n".join(merge_commands + move_ops)
        if full_cmd:
            # write full_cmd to a .sh file
            # Open file for writing
            fileObject = open(new_prefix + "_full_cmd.sh", "w")
            fileObject.write("#!/bin/bash\n")
            fileObject.write(full_cmd)
            # Close the file
            fileObject.close()

            renames = new_prefix + '_full_cmd.sh'

            if self.use_datalad:

                # first check if IntendedFor renames need to be saved
                if not self.is_datalad_clean():
                    s1 = "Renamed IntendedFor references to "
                    s2 = "Variant Group scans"
                    IF_rename_msg = s1 + s2
                    self.datalad_handle.save(message=IF_rename_msg)

                s1 = "Renamed Variant Group scans according to their variant "
                s2 = "parameters"

                rename_commit = s1 + s2

                self.datalad_handle.run(cmd=["bash", renames],
                                        message=rename_commit)
            else:
                subprocess.run(["bash", renames],
                               stdout=subprocess.PIPE,
                               cwd=str(Path(new_prefix).parent))
        else:
            print("Not running any commands")

        self.reset_bids_layout()
        self.get_CSVs(new_prefix)

    def change_filename(self, filepath, entities):
        """Applies changes to a filename based on the renamed
        key groups.
        This function takes into account the new key group names
        and renames all files whose key group names changed.
        Parameters:
        -----------
            filepath : str
                Path prefix to a file in the affected key group change
            entities : dictionary
                A pybids dictionary of entities parsed from the new key
                group name.
        """
        exts = Path(filepath).suffixes
        old_ext = ""
        for ext in exts:
            old_ext += ext

        suffix = entities['suffix']
        entity_file_keys = []
        file_keys = ['task', 'acquisition', 'direction',
                     'reconstruction', 'run']

        for key in file_keys:
            if key in list(entities.keys()):
                entity_file_keys.append(key)

        sub = get_key_name(filepath, 'sub')
        ses = get_key_name(filepath, 'ses')
        sub_ses = sub + '_' + ses

        if 'run' in list(entities.keys()) and 'run-0' in filepath:
            entities['run'] = '0' + str(entities['run'])

        filename = "_".join(["{}-{}".format(key, entities[key])
                             for key in entity_file_keys])
        filename = filename.replace('acquisition', 'acq') \
            .replace('direction', 'dir') \
            .replace('reconstruction', 'rec')
        if len(filename) > 0:
            filename = sub_ses + '_' + filename + '_' + suffix + old_ext
        else:
            filename = sub_ses + filename + '_' + suffix + old_ext

        # CHECK TO SEE IF DATATYPE CHANGED
        dtypes = ['anat', 'func', 'perf', 'fmap', 'dwi']
        old = ''
        for dtype in dtypes:
            if dtype in filepath:
                old = dtype

        if 'datatype' in entities.keys():
            dtype = entities['datatype']
            if entities['datatype'] != old:
                print("WARNING: DATATYPE CHANGE DETECETD")
        else:
            dtype = old
        new_path = str(self.path) + '/' + sub + '/' + ses \
            + '/' + dtype + '/' + filename

        # add the scan path + new path to the lists of old, new filenames
        self.old_filenames.append(filepath)
        self.new_filenames.append(new_path)

        # NOW NEED TO RENAME ASSOCIATIONS
        bids_file = self.layout.get_file(filepath)
        associations = bids_file.get_associations()
        for assoc in associations:
            assoc_path = assoc.path
            if Path(assoc_path).exists():
                # print("FILE: ", filepath)
                # print("ASSOC: ", assoc.path)
                # ensure assoc not an IntendedFor reference
                if '.nii' not in str(assoc_path):
                    self.old_filenames.append(assoc_path)
                    new_ext_path = img_to_new_ext(new_path,
                                                  ''.join(Path(assoc.path)
                                                          .suffixes))
                    self.new_filenames.append(new_ext_path)

        # MAKE SURE THESE AREN'T COVERED BY get_associations!!!
        if '/dwi/' in filepath:
            # add the bval and bvec if there
            if Path(img_to_new_ext(filepath, '.bval')).exists() \
                    and img_to_new_ext(filepath, '.bval') \
                    not in self.old_filenames:
                self.old_filenames.append(img_to_new_ext(filepath,
                                                         '.bval'))
                self.new_filenames.append(img_to_new_ext(new_path,
                                                         '.bval'))

            if Path(img_to_new_ext(filepath, '.bvec')).exists() \
                    and img_to_new_ext(filepath, '.bvec') \
                    not in self.old_filenames:
                self.old_filenames.append(img_to_new_ext(filepath,
                                                         '.bvec'))
                self.new_filenames.append(img_to_new_ext(new_path,
                                                         '.bvec'))

        # now rename _events and _physio files!
        old_suffix = parse_file_entities(filepath)['suffix']
        scan_end = '_' + old_suffix + old_ext

        if '_task-' in filepath:
            old_events = filepath.replace(scan_end, '_events.tsv')
            old_ejson = filepath.replace(scan_end, '_events.json')
            if Path(old_events).exists():
                self.old_filenames.append(old_events)
                new_scan_end = '_' + suffix + old_ext
                new_events = new_path.replace(new_scan_end, '_events.tsv')
                self.new_filenames.append(new_events)
            if Path(old_ejson).exists():
                self.old_filenames.append(old_ejson)
                new_scan_end = '_' + suffix + old_ext
                new_ejson = new_path.replace(new_scan_end, '_events.json')
                self.new_filenames.append(new_ejson)

        old_physio = filepath.replace(scan_end, '_physio.tsv.gz')
        if Path(old_physio).exists():
            self.old_filenames.append(old_physio)
            new_scan_end = '_' + suffix + old_ext
            new_physio = new_path.replace(new_scan_end, '_physio.tsv.gz')
            self.new_filenames.append(new_physio)

        # RENAME INTENDED FORS!
        ses_path = self.path + '/' + sub + '/' + ses
        for path in Path(ses_path).rglob("fmap/*.json"):
            self.IF_rename_paths.append(str(path))
            json_file = self.layout.get_file(str(path))
            data = json_file.get_dict()

            if 'IntendedFor' in data.keys():
                # check if IntendedFor field is a str or list
                if isinstance(data['IntendedFor'], str):
                    if data['IntendedFor'] == \
                            _get_intended_for_reference(filepath):
                        # replace old filename with new one (overwrite string)
                        data['IntendedFor'] = \
                                _get_intended_for_reference(new_path)

                        # update the json with the new data dictionary
                        _update_json(json_file.path, data)

                if isinstance(data['IntendedFor'], list):
                    for item in data['IntendedFor']:
                        if item in _get_intended_for_reference(filepath):

                            # remove old filename
                            data['IntendedFor'].remove(item)
                            # add new filename
                            data['IntendedFor'].append(
                                    _get_intended_for_reference(new_path))

                        # update the json with the new data dictionary
                        _update_json(json_file.path, data)

        # save IntendedFor purges so that you can datalad run the
        # remove association file commands on a clean dataset
        # if self.use_datalad:
        #     if not self.is_datalad_clean():
        #         self.datalad_save(message="Renamed IntendedFors")
        #         self.reset_bids_layout()
            # else:
            #     print("No IntendedFor References to Rename")

    def copy_exemplars(self, exemplars_dir, exemplars_csv, min_group_size,
                       raise_on_error=True):
        """Copies one subject from each Acquisition Group into a new directory
        for testing *preps, raises an error if the subjects are not unlocked,
        unlocks each subject before copying if --force_unlock is set.

        Parameters:
        -----------
            exemplars_dir: str
                path to the directory that will contain one subject
                from each Acqusition Gorup (*_AcqGrouping.csv)
                example path: /Users/Covitz/CSVs/CCNP_Acq_Groups/

            exemplars_csv: str
                path to the .csv file that lists one subject
                from each Acqusition Group (*_AcqGrouping.csv
                from the cubids-group output)
                example path: /Users/Covitz/CSVs/CCNP_Acq_Grouping.csv
        """
        # create the exemplar ds
        if self.use_datalad:
            subprocess.run(['datalad', '--log-level', 'error', 'create', '-c',
                            'text2git', exemplars_dir])

        # load the exemplars csv
        subs = pd.read_csv(exemplars_csv)

        # if min group size flag set, drop acq groups with less than min
        if int(min_group_size) > 1:
            for row in range(len(subs)):
                acq_group = subs.loc[row, 'AcqGroup']
                size = int(subs['AcqGroup'].value_counts()[acq_group])
                if size < int(min_group_size):
                    subs = subs.drop([row])

        # get one sub from each acq group
        unique = subs.drop_duplicates(subset=["AcqGroup"])

        # cast list to a set to drop duplicates, then convert back to list
        unique_subs = list(set(unique['subject'].tolist()))
        for subid in unique_subs:
            source = str(self.path) + '/' + subid
            dest = exemplars_dir + '/' + subid
            # Copy the content of source to destination
            copytree(source, dest)

        # Copy the dataset_description.json
        copyfile(str(self.path) + '/' + 'dataset_description.json',
                 exemplars_dir + '/' + 'dataset_description.json')

        s1 = "Copied one subject from each Acquisition Group "
        s2 = "into the Exemplar Dataset"
        msg = s1 + s2
        if self.use_datalad:
            subprocess.run(['datalad', 'save', '-d', exemplars_dir,
                            '-m', msg])

    def purge(self, scans_txt, raise_on_error=True):
        """Purges all associations of desired scans from a bids dataset.

        Parameters:
        -----------
            scans_txt: str
                path to the .txt file that lists the scans
                you want to be deleted from the dataset, along
                with thier associations.
                example path: /Users/Covitz/CCNP/scans_to_delete.txt
        """

        self.scans_txt = scans_txt

        scans = []
        with open(scans_txt, 'r') as fd:
            reader = csv.reader(fd)
            for row in reader:
                scans.append(str(row[0]))

        # check to ensure scans are all real files in the ds!

        self._purge_associations(scans)

    def _purge_associations(self, scans):

        # PURGE FMAP JSONS' INTENDED FOR REFERENCES

        # truncate all paths to intendedfor reference format
        # sub, ses, modality only (no self.path)
        if_scans = []
        for scan in scans:
            if_scans.append(_get_intended_for_reference(scan))

        for path in Path(self.path).rglob("sub-*/*/fmap/*.json"):

            json_file = self.layout.get_file(str(path))

            data = json_file.get_dict()

            # remove scan references in the IntendedFor

            if 'IntendedFor' in data.keys():
                # check if IntendedFor field value is a list or a string
                if isinstance(data['IntendedFor'], str):
                    if data['IntendedFor'] in if_scans:
                        data['IntendedFor'] = []
                        # update the json with the new data dictionary
                        _update_json(json_file.path, data)

                if isinstance(data['IntendedFor'], list):
                    for item in data['IntendedFor']:
                        if item in if_scans:
                            data['IntendedFor'].remove(item)

                            # update the json with the new data dictionary
                            _update_json(json_file.path, data)

        # save IntendedFor purges so that you can datalad run the
        # remove association file commands on a clean dataset
        if self.use_datalad:
            if not self.is_datalad_clean():
                s1 = "Purged IntendedFor references to files "
                s2 = "requested for removal"
                message = s1 + s2
                self.datalad_save(message=message)
                self.reset_bids_layout()

        # NOW WE WANT TO PURGE ALL ASSOCIATIONS

        to_remove = []

        for path in Path(self.path).rglob("sub-*/**/*.nii.gz"):

            if str(path) in scans:
                bids_file = self.layout.get_file(str(path))
                associations = bids_file.get_associations()

                for assoc in associations:
                    filepath = assoc.path

                    # ensure association is not an IntendedFor reference!
                    if '.nii' not in str(filepath):
                        to_remove.append(filepath)
                if '/dwi/' in str(path):
                    # add the bval and bvec if there
                    if Path(img_to_new_ext(str(path), '.bval')).exists():
                        to_remove.append(img_to_new_ext(str(path), '.bval'))
                    if Path(img_to_new_ext(str(path), '.bvec')).exists():
                        to_remove.append(img_to_new_ext(str(path), '.bvec'))
                if '/func/' in str(path):
                    # add tsvs
                    tsv = img_to_new_ext(str(path), '.tsv').replace(
                            '_bold', '_events')
                    if Path(tsv).exists():
                        to_remove.append(tsv)
                    # add tsv json (if exists)
                    if Path(tsv.replace('.tsv', '.json')).exists():
                        to_remove.append(tsv.replace('.tsv', '.json'))
        to_remove += scans

        # create rm commands for all files that need to be purged
        purge_commands = []
        for rm_me in to_remove:
            if Path(rm_me).exists():
                purge_commands.append("rm " + rm_me)

        # datalad run the file deletions (purges)
        full_cmd = "\n".join(purge_commands)
        if full_cmd:

            # write full_cmd to a .sh file
            # Open file for writing

            path_prefix = str(Path(self.path).parent)

            fileObject = open(path_prefix + "/" + "_full_cmd.sh", "w")
            fileObject.write("#!/bin/bash\n")
            fileObject.write(full_cmd)
            # Close the file
            fileObject.close()
            if self.scans_txt:
                cmt = "Purged scans listed in %s from dataset" % self.scans_txt
            else:
                cmt = "Purged Parameter Groups marked for removal"
            purge_file = path_prefix + "/" + '_full_cmd.sh'
            if self.use_datalad:
                self.datalad_handle.run(cmd=["bash", purge_file],
                                        message=cmt)
            else:
                subprocess.run(["bash", path_prefix + "/" + "_full_cmd.sh"],
                               stdout=subprocess.PIPE,
                               cwd=path_prefix)
            self.reset_bids_layout()
        else:
            print("Not running any association removals")

    def _cache_fieldmaps(self):
        """Searches all fieldmaps and creates a lookup for each file.

        Returns:
        -----------
            misfits : list
                A list of fmap filenames for whom CuBIDS has not detected
                an IntnededFor.
        """

        suffix = '(phase1|phasediff|epi|fieldmap)'
        fmap_files = self.layout.get(suffix=suffix, regex_search=True,
                                     extension=['.nii.gz', '.nii'])
        misfits = []
        files_to_fmaps = defaultdict(list)
        for fmap_file in tqdm(fmap_files):
            intentions = listify(fmap_file.get_metadata().get("IntendedFor"))
            subject_prefix = "sub-%s" % fmap_file.entities['subject']

            if intentions is not None:
                for intended_for in intentions:
                    full_path = Path(self.path) / subject_prefix / intended_for
                    files_to_fmaps[str(full_path)].append(fmap_file)

            # fmap file detected, no intended for found
            else:
                misfits.append(fmap_file)

        self.fieldmap_lookup = files_to_fmaps
        self.fieldmaps_cached = True

        # return a list of all filenames where fmap file detected,
        # no intended for found
        return misfits

    def get_param_groups_from_key_group(self, key_group):
        """Splits key groups into param groups based on json metadata.

        Parameters:
        -----------
            key_group : str
                Key group name.

        Returns:
        -----------
            ret : tuple of two DataFrames
                1. A data frame with one row per file where the ParamGroup
                column indicates the group to which each scan belongs.
                2. A data frame with param group summaries
        """
        if not self.fieldmaps_cached:
            raise Exception(
                "Fieldmaps must be cached to find parameter groups.")
        key_entities = _key_group_to_entities(key_group)
        key_entities["extension"] = ".nii[.gz]*"

        matching_files = self.layout.get(return_type="file", scope="self",
                                         regex_search=True, **key_entities)

        # ensure files who's entities contain key_entities but include other
        # entities do not also get added to matching_files
        to_include = []
        for filepath in matching_files:
            f_key_group = _file_to_key_group(filepath)

            if f_key_group == key_group:
                to_include.append(filepath)

        # get the modality associated with the key group
        modalities = ['/dwi/', '/anat/', '/func/', '/perf/', '/fmap/']
        modality = ''
        for mod in modalities:
            if mod in filepath:
                modality = mod.replace('/', '').replace('/', '')
        if modality == '':
            print("Unusual Modality Detected")
            modality = 'other'

        ret = _get_param_groups(
            to_include, self.layout, self.fieldmap_lookup, key_group,
            self.grouping_config, modality, self.keys_files)

        # add modality to the retun tuple
        l_ret = list(ret)
        l_ret.append(modality)
        tup_ret = tuple(l_ret)
        return tup_ret

    def get_param_groups_dataframes(self):
        '''Creates DataFrames of files x param groups and a summary'''

        key_groups = self.get_key_groups()
        labeled_files = []
        param_group_summaries = []
        for key_group in key_groups:
            labeled_file_params, param_summary, modality = \
                self.get_param_groups_from_key_group(key_group)
            if labeled_file_params is None:
                continue
            param_group_summaries.append(param_summary)
            labeled_files.append(labeled_file_params)

        big_df = _order_columns(pd.concat(labeled_files, ignore_index=True))
        summary = _order_columns(pd.concat(param_group_summaries,
                                 ignore_index=True))

        # create new col that strings key and param group together
        summary["KeyParamGroup"] = summary["KeyGroup"] \
            + '__' + summary["ParamGroup"].map(str)

        # move this column to the front of the dataframe
        key_param_col = summary.pop("KeyParamGroup")
        summary.insert(0, "KeyParamGroup", key_param_col)

        # do the same for the files df
        big_df["KeyParamGroup"] = big_df["KeyGroup"] \
            + '__' + big_df["ParamGroup"].map(str)

        # move this column to the front of the dataframe
        key_param_col = big_df.pop("KeyParamGroup")
        big_df.insert(0, "KeyParamGroup", key_param_col)

        summary.insert(0, "RenameKeyGroup", np.nan)
        summary.insert(0, "MergeInto", np.nan)
        summary.insert(0, "ManualCheck", np.nan)
        summary.insert(0, "Notes", np.nan)

        # NOW WANT TO AUTOMATE RENAME!
        # loop though imaging and derrived param keys

        sidecar = self.grouping_config.get('sidecar_params')
        sidecar = sidecar[modality]

        relational = self.grouping_config.get('relational_params')

        # list of columns names that we account for in suggested renaming
        summary['RenameKeyGroup'] = summary['RenameKeyGroup'].apply(str)

        rename_cols = []
        tolerance_cols = []
        for col in sidecar.keys():
            if 'suggest_variant_rename' in sidecar[col].keys():
                if sidecar[col]['suggest_variant_rename'] \
                        and col in summary.columns:
                    rename_cols.append(col)
                    if 'tolerance' in sidecar[col].keys():
                        tolerance_cols.append(col)

        # deal with Fmap!
        if 'FieldmapKey' in relational:
            if 'suggest_variant_rename' in relational['FieldmapKey'].keys():
                if relational['FieldmapKey']['suggest_variant_rename']:
                    # check if 'bool' or 'columns'
                    if relational['FieldmapKey']['display_mode'] == 'bool':
                        rename_cols.append("HasFieldmap")

        # deal with IntendedFor Key!
        if 'IntendedForKey' in relational:
            if 'suggest_variant_rename' in relational['IntendedForKey'].keys():
                if relational['FieldmapKey']['suggest_variant_rename']:
                    # check if 'bool' or 'columns'
                    if relational['IntendedForKey']['display_mode'] == 'bool':
                        rename_cols.append("UsedAsFieldmap")

        dom_dict = {}
        # loop through summary csv and create dom_dict
        for row in range(len(summary)):
            # if 'NumVolumes' in summary.columns \
            #         and str(summary.loc[row, "NumVolumes"]) == 'nan':
            #     summary.at[row, "NumVolumes"] = 1.0

            # if dominant group identified
            if str(summary.loc[row, 'ParamGroup']) == '1':
                val = {}
                # grab col, all vals send to dict
                key = summary.loc[row, "KeyGroup"]
                for col in rename_cols:
                    summary[col] = summary[col].apply(str)
                    val[col] = summary.loc[row, col]
                dom_dict[key] = val

        # now loop through again and ID variance
        for row in range(len(summary)):
            # check to see if renaming has already happened
            renamed = False
            entities = _key_group_to_entities(summary.loc[row, "KeyGroup"])
            if 'VARIANT' in summary.loc[row, 'KeyGroup']:
                renamed = True

            # if NumVolumes is nan, set to 1.0
            # if 'NumVolumes' in summary.columns \
            #         and str(summary.loc[row, "NumVolumes"]) == 'nan':
            #     summary.at[row, "NumVolumes"] = 1.0

            if summary.loc[row, "ParamGroup"] != 1 and not renamed:
                acq_str = 'VARIANT'
                # now we know we have a deviant param group
                # check if TR is same as param group 1
                key = summary.loc[row, "KeyGroup"]
                for col in rename_cols:
                    summary[col] = summary[col].apply(str)
                    if summary.loc[row, col] != dom_dict[key][col]:

                        if col == 'HasFieldmap':
                            if dom_dict[key][col] == 'True':
                                acq_str = acq_str + 'NoFmap'
                            else:
                                acq_str = acq_str + 'HasFmap'
                        elif col == 'UsedAsFieldmap':
                            if dom_dict[key][col] == 'True':
                                acq_str = acq_str + 'Unused'
                            else:
                                acq_str = acq_str + 'IsUsed'
                        else:
                            acq_str = acq_str + col

                if acq_str == 'VARIANT':
                    acq_str = acq_str + 'Other'

                if 'acquisition' in entities.keys():
                    acq = 'acquisition-%s' % entities['acquisition'] + acq_str

                    new_name = summary.loc[row, "KeyGroup"].replace(
                            'acquisition-%s' % entities['acquisition'], acq)
                else:
                    acq = 'acquisition-%s' % acq_str
                    new_name = acq + '_' + summary.loc[row, "KeyGroup"]

                summary.at[row, 'RenameKeyGroup'] = new_name

            # convert all "nan" to empty str
            # so they don't show up in the summary csv
            if summary.loc[row, "RenameKeyGroup"] == 'nan':
                summary.at[row, "RenameKeyGroup"] = ''

            for col in rename_cols:
                if summary.loc[row, col] == 'nan':
                    summary.at[row, col] = ''

        return (big_df, summary)

    def get_CSVs(self, path_prefix):
        """Creates the _summary and _files CSVs for the bids dataset.

        Parameters:
        -----------
            prefix_path: str
                prefix of the path to the directory where you want
                to save your CSVs
                example path: /Users/Covitz/PennLINC/RBC/CCNP/
        """

        self._cache_fieldmaps()

        big_df, summary = self.get_param_groups_dataframes()

        summary = summary.sort_values(by=['Modality', 'KeyGroupCount'],
                                      ascending=[True, False])
        big_df = big_df.sort_values(by=['Modality', 'KeyGroupCount'],
                                    ascending=[True, False])

        big_df.to_csv(path_prefix + "_files.csv", index=False)
        summary.to_csv(path_prefix + "_summary.csv", index=False)

        # Calculate the acq groups
        group_by_acquisition_sets(path_prefix + "_files.csv", path_prefix,
                                  self.acq_group_level)

        print("Detected " + str(len(summary)) + " Parameter Groups.")

    def get_key_groups(self):
        '''Identifies the key groups for the bids dataset'''

        # reset self.keys_files
        self.keys_files = {}

        key_groups = set()

        for path in Path(self.path).rglob("sub-*/**/*.*"):
            # ignore all dot directories
            if '/.' in str(path):
                continue

            if str(path).endswith(".nii") or str(path).endswith(".nii.gz"):
                key_groups.update((_file_to_key_group(path),))

                # Fill the dictionary of key group, list of filenames pairrs
                ret = _file_to_key_group(path)

                if ret not in self.keys_files.keys():

                    self.keys_files[ret] = []

                self.keys_files[ret].append(path)

        return sorted(key_groups)

    def change_metadata(self, filters, pattern, metadata):

        files_to_change = self.layout.get(return_type='object', **filters)

        for bidsfile in files_to_change:
            # get the sidecar file
            bidsjson_file = bidsfile.get_associations()

            if not bidsjson_file:
                print("NO JSON FILES FOUND IN ASSOCIATIONS")
                continue

            json_file = [x for x in bidsjson_file if 'json' in x.filename]
            if not len(json_file) == 1:
                print("FOUND IRREGULAR ASSOCIATIONS")

            else:
                # get the data from it
                json_file = json_file[0]

                sidecar = json_file.get_dict()
                sidecar.update(metadata)

                # write out
                _update_json(json_file.path, sidecar)

    def get_all_metadata_fields(self):
        ''' Returns all metadata fields in a bids directory'''

        found_fields = set()
        for json_file in Path(self.path).rglob("*.json"):
            if '.git' not in str(json_file):
                with open(json_file, "r") as jsonr:
                    metadata = json.load(jsonr)
                found_fields.update(metadata.keys())
        return sorted(found_fields)

    def remove_metadata_fields(self, fields_to_remove):
        '''Removes specific fields from all metadata files.'''

        remove_fields = set(fields_to_remove)
        if not remove_fields:
            return
        for json_file in tqdm(Path(self.path).rglob("*.json")):
            # Check for offending keys in the json file
            if '.git' not in str(json_file):
                with open(json_file, "r") as jsonr:
                    metadata = json.load(jsonr)
                offending_keys = remove_fields.intersection(metadata.keys())
                # Quit if there are none in there
                if not offending_keys:
                    continue

                # Remove the offending keys
                for key in offending_keys:
                    del metadata[key]
                # Write the cleaned output
                with open(json_file, "w") as jsonr:
                    json.dump(metadata, jsonr, indent=4)

    # # # # FOR TESTING # # # #
    def get_filenames(self):
        return self.keys_files

    def get_fieldmap_lookup(self):
        return self.fieldmap_lookup

    def get_layout(self):
        return self.layout


def _validateJSON(json_file):
    # TODO: implement this or delete ???
    return True


def _update_json(json_file, metadata):

    if _validateJSON(metadata):
        with open(json_file, 'w', encoding='utf-8') as f:
            json.dump(metadata, f, ensure_ascii=False, indent=4)
    else:
        print("INVALID JSON DATA")


def _key_group_to_entities(key_group):
    '''Splits a key_group name into a pybids dictionary of entities.'''

    return dict([group.split("-") for group in key_group.split("_")])


def _entities_to_key_group(entities):
    '''Converts a pybids entities dictionary into a key group name.'''

    group_keys = sorted(entities.keys() - NON_KEY_ENTITIES)
    return "_".join(
        ["{}-{}".format(key, entities[key]) for key in group_keys])


def _file_to_key_group(filename):
    '''Identifies and returns the key group of a bids valid filename.'''

    entities = parse_file_entities(str(filename))
    return _entities_to_key_group(entities)


def _get_intended_for_reference(scan):
    return '/'.join(Path(scan).parts[-3:])


def _get_param_groups(files, layout, fieldmap_lookup, key_group_name,
                      grouping_config, modality, keys_files):

    """Finds a list of *parameter groups* from a list of files.

    For each file in `files`, find critical parameters for metadata. Then find
    unique sets of these critical parameters.

    Parameters:
    -----------
    files : list
        List of file names

    layout : bids.BIDSLayout
        PyBIDS BIDSLayout object where `files` come from

    fieldmap_lookup : defaultdict
        mapping of filename strings relative to the bids root
        (e.g. "sub-X/ses-Y/func/sub-X_ses-Y_task-rest_bold.nii.gz")

    grouping_config : dict
        configuration for defining parameter groups

    Returns:
    --------
    labeled_files : pd.DataFrame
        A data frame with one row per file where the ParamGroup column
        indicates which group each scan is a part of.

    param_groups_with_counts : pd.DataFrame
        A data frame with param group summaries

    """

    if not files:
        print("WARNING: no files for", key_group_name)
        return None, None

    # Split the config into separate parts
    imaging_params = grouping_config.get('sidecar_params', {})
    imaging_params = imaging_params[modality]

    relational_params = grouping_config.get('relational_params', {})

    derived_params = grouping_config.get('derived_params')
    derived_params = derived_params[modality]

    imaging_params.update(derived_params)

    dfs = []
    # path needs to be relative to the root with no leading prefix
    for path in files:
        metadata = layout.get_metadata(path)
        intentions = metadata.get("IntendedFor", [])
        slice_times = metadata.get("SliceTiming", [])

        wanted_keys = metadata.keys() & imaging_params
        example_data = {key: metadata[key] for key in wanted_keys}
        example_data["KeyGroup"] = key_group_name

        # Get the fieldmaps out and add their types
        if 'FieldmapKey' in relational_params:
            fieldmap_types = sorted([_file_to_key_group(fmap.path) for
                                    fmap in fieldmap_lookup[path]])

            # check if config says columns or bool
            if relational_params['FieldmapKey']['display_mode'] == \
                    'bool':
                if len(fieldmap_types) > 0:
                    example_data['HasFieldmap'] = True
                else:
                    example_data['HasFieldmap'] = False
            else:
                for fmap_num, fmap_type in enumerate(fieldmap_types):
                    example_data['FieldmapKey%02d' % fmap_num] = fmap_type

        # Add the number of slice times specified
        if "NSliceTimes" in derived_params:
            example_data["NSliceTimes"] = len(slice_times)

        example_data["FilePath"] = path

        # If it's a fieldmap, see what key group it's intended to correct
        if "IntendedForKey" in relational_params:
            intended_key_groups = sorted([_file_to_key_group(intention) for
                                          intention in intentions])

            # check if config says columns or bool
            if relational_params['IntendedForKey']['display_mode'] == \
                    'bool':
                if len(intended_key_groups) > 0:
                    example_data["UsedAsFieldmap"] = True
                else:
                    example_data["UsedAsFieldmap"] = False
            else:
                for intention_num, intention_key_group in \
                        enumerate(intended_key_groups):
                    example_data[
                        "IntendedForKey%02d" % intention_num] = \
                                intention_key_group

        dfs.append(example_data)

    # Assign each file to a ParamGroup

    # round param groups based on precision
    df = round_params(pd.DataFrame(dfs), grouping_config, modality)

    # cluster param groups based on tolerance
    df = format_params(df, grouping_config, modality)
    # param_group_cols = list(set(df.columns.to_list()) - set(["FilePath"]))

    # get the subset of columns to drop duplicates by
    check_cols = []
    for col in list(df.columns):
        if "Cluster_" + col not in list(df.columns) and col != 'FilePath':
            check_cols.append(col)

    # Find the unique ParamGroups and assign ID numbers in "ParamGroup"
    deduped = df.drop('FilePath', axis=1).drop_duplicates(subset=check_cols,
                                                          ignore_index=True)

    deduped["ParamGroup"] = np.arange(deduped.shape[0]) + 1

    # add the modality as a column
    deduped["Modality"] = modality

    # add key group count column (will delete later)
    deduped["KeyGroupCount"] = len(keys_files[key_group_name])

    # Add the ParamGroup to the whole list of files
    labeled_files = pd.merge(df, deduped, on=check_cols)

    value_counts = labeled_files.ParamGroup.value_counts()

    param_group_counts = pd.DataFrame(
        {"Counts": value_counts.to_numpy(),
         "ParamGroup": value_counts.index.to_numpy()})

    param_groups_with_counts = pd.merge(
        deduped, param_group_counts, on=["ParamGroup"])

    # Sort by counts and relabel the param groups
    param_groups_with_counts.sort_values(by=['Counts'], inplace=True,
                                         ascending=False)
    param_groups_with_counts["ParamGroup"] = np.arange(
        param_groups_with_counts.shape[0]) + 1

    # Send the new, ordered param group ids to the files list
    ordered_labeled_files = pd.merge(df, param_groups_with_counts,
                                     on=check_cols, suffixes=('_x', ''))

    # sort ordered_labeled_files by param group
    ordered_labeled_files.sort_values(by=['Counts'], inplace=True,
                                      ascending=False)

    # now get rid of cluster cols from deduped and df
    for col in list(ordered_labeled_files.columns):
        if col.startswith('Cluster_'):
            ordered_labeled_files = ordered_labeled_files.drop(col, axis=1)
            param_groups_with_counts = param_groups_with_counts.drop(col,
                                                                     axis=1)
        if col.endswith('_x'):
            ordered_labeled_files = ordered_labeled_files.drop(col, axis=1)

    return ordered_labeled_files, param_groups_with_counts


def round_params(param_group_df, config, modality):
    to_format = config['sidecar_params'][modality]
    to_format.update(config['derived_params'][modality])

    for column_name, column_fmt in to_format.items():
        if column_name not in param_group_df:
            continue
        if 'precision' in column_fmt:
            param_group_df[column_name] = \
                param_group_df[column_name].round(column_fmt['precision'])

    return param_group_df


def format_params(param_group_df, config, modality):
    '''Run AgglomerativeClustering on param groups, add columns to dataframe'''

    to_format = config['sidecar_params'][modality]
    to_format.update(config['derived_params'][modality])

    for column_name, column_fmt in to_format.items():
        if column_name not in param_group_df:
            continue
        if 'tolerance' in column_fmt and len(param_group_df) > 1:
            array = param_group_df[column_name].to_numpy().reshape(-1, 1)

            for i in range(len(array)):
                if np.isnan(array[i, 0]):
                    array[i, 0] = -999

            tolerance = to_format[column_name]['tolerance']
            clustering = AgglomerativeClustering(n_clusters=None,
                                                 distance_threshold=tolerance,
                                                 linkage='complete').fit(array)
            for i in range(len(array)):
                if array[i, 0] == -999:
                    array[i, 0] = np.nan

            # now add clustering_labels as a column
            param_group_df['Cluster_' + column_name] = clustering.labels_

    return param_group_df


def _order_columns(df):
    '''Organizes columns of the summary and files DataFrames so that
    KeyGroup and ParamGroup are the first two columns, FilePath is
    the last, and the others are sorted alphabetically.'''

    cols = set(df.columns.to_list())
    non_id_cols = cols - ID_VARS
    new_columns = ["KeyGroup", "ParamGroup"] + sorted(non_id_cols)
    if "FilePath" in cols:
        new_columns.append("FilePath")

    df = df[new_columns]

    return df[new_columns]


def img_to_new_ext(img_path, new_ext):
    # handle .tsv edge case
    if new_ext == '.tsv':
        # take out suffix
        return img_path.rpartition('_')[0] + '_events' + new_ext
    if new_ext == '.tsv.gz':
        return img_path.rpartition('_')[0] + '_physio' + new_ext
    else:
        return img_path.replace(".nii.gz", "").replace(".nii", "") + new_ext


def get_key_name(path, key):
    # given a filepath and BIDS key name, return value
    parts = Path(path).parts
    for part in parts:
        if part.startswith(key + '-'):
            return part



"""Console script for cubids."""
import warnings
import argparse
import subprocess
import os
import sys
import re
import logging
import tempfile
import tqdm
import shutil
import pandas as pd
from cubids import CuBIDS
from pathlib import Path
from .validator import (build_validator_call,
                        run_validator, parse_validator_output,
                        build_subject_paths)
from .metadata_merge import merge_json_into_json

warnings.simplefilter(action='ignore', category=FutureWarning)
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('cubids-cli')
GIT_CONFIG = os.path.join(os.path.expanduser("~"), '.gitconfig')
logging.getLogger('datalad').setLevel(logging. ERROR)


def cubids_validate():
    '''Command Line Interface function for running the bids validator.'''

    parser = argparse.ArgumentParser(
        description="cubids-validate: Wrapper around the official "
        "BIDS Validator",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='the root of a BIDS dataset. It should contain '
                        'sub-X directories and dataset_description.json')
    parser.add_argument('output_prefix',
                        type=Path,
                        action='store',
                        help='file prefix to which tabulated validator output '
                        'is written.')
    parser.add_argument('--sequential',
                        action='store_true',
                        default=False,
                        help='Run the BIDS validator sequentially '
                        'on each subject.',
                        required=False)
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.',
                        default=None)
    parser.add_argument('--ignore_nifti_headers',
                        action='store_true',
                        default=False,
                        help='Disregard NIfTI header content during'
                        ' validation',
                        required=False)
    parser.add_argument('--ignore_subject_consistency',
                        action='store_true',
                        default=True,
                        help='Skip checking that any given file for one'
                        ' subject is present for all other subjects',
                        required=False)
    parser.add_argument('--sequential-subjects',
                        action='store',
                        default=None,
                        help='List: Filter the sequential run to only include'
                        ' the listed subjects. e.g. --sequential-subjects '
                        'sub-01 sub-02 sub-03',
                        nargs='+',
                        required=False)
    opts = parser.parse_args()

    # Run directly from python using subprocess
    if opts.container is None:

        if not opts.sequential:
            # run on full dataset
            call = build_validator_call(str(opts.bids_dir),
                                        opts.ignore_nifti_headers,
                                        opts.ignore_subject_consistency)
            ret = run_validator(call)

            # parse the string output
            parsed = parse_validator_output(ret.stdout.decode('UTF-8'))
            if parsed.shape[1] < 1:
                logger.info("No issues/warnings parsed, your dataset"
                            " is BIDS valid.")
                sys.exit(0)
            else:
                logger.info("BIDS issues/warnings found in the dataset")

                if opts.output_prefix:
                    # normally, write dataframe to file in CLI
                    val_csv = str(opts.output_prefix) + "_validation.csv"
                    parsed.to_csv(val_csv, index=False)
                    logger.info("Writing issues out to %s", val_csv)
                    sys.exit(0)
                else:
                    # user may be in python session, return dataframe
                    return parsed
        else:
            # logger.info("Prepping sequential validator run...")

            # build a dictionary with {SubjectLabel: [List of files]}
            subjects_dict = build_subject_paths(opts.bids_dir)

            # logger.info("Running validator sequentially...")
            # iterate over the dictionary

            parsed = []

            if opts.sequential_subjects:
                subjects_dict = {k: v for k, v in subjects_dict.items()
                                 if k in opts.sequential_subjects}
            assert len(list(subjects_dict.keys())) > 1, ("No subjects found"
                                                         " in filter")
            for subject, files_list in tqdm.tqdm(subjects_dict.items()):

                # logger.info(" ".join(["Processing subject:", subject]))
                # create a temporary directory and symlink the data
                with tempfile.TemporaryDirectory() as tmpdirname:
                    for fi in files_list:

                        # cut the path down to the subject label
                        bids_start = fi.find(subject)

                        # maybe it's a single file
                        if bids_start < 1:
                            bids_folder = tmpdirname
                            fi_tmpdir = tmpdirname

                        else:
                            bids_folder = Path(fi[bids_start:]).parent
                            fi_tmpdir = tmpdirname + '/' + str(bids_folder)

                        if not os.path.exists(fi_tmpdir):
                            os.makedirs(fi_tmpdir)
                        output = fi_tmpdir + '/' + str(Path(fi).name)
                        shutil.copy2(fi, output)

                    # run the validator
                    nifti_head = opts.ignore_nifti_headers
                    subj_consist = opts.ignore_subject_consistency
                    call = build_validator_call(tmpdirname,
                                                nifti_head,
                                                subj_consist)
                    ret = run_validator(call)
                    # parse output
                    if ret.returncode != 0:
                        logger.error("Errors returned "
                                     "from validator run, parsing now")

                    # parse the output and add to list if it returns a df
                    decoded = ret.stdout.decode('UTF-8')
                    tmp_parse = parse_validator_output(decoded)
                    if tmp_parse.shape[1] > 1:
                        tmp_parse['subject'] = subject
                        parsed.append(tmp_parse)

            # concatenate the parsed data and exit
            if len(parsed) < 1:
                logger.info("No issues/warnings parsed, your dataset"
                            " is BIDS valid.")
                sys.exit(0)

            else:
                parsed = pd.concat(parsed, axis=0)
                subset = parsed.columns.difference(['subject'])
                parsed = parsed.drop_duplicates(subset=subset)

                logger.info("BIDS issues/warnings found in the dataset")

                if opts.output_prefix:
                    # normally, write dataframe to file in CLI
                    val_csv = str(opts.output_prefix) + "_validation.csv"
                    parsed.to_csv(val_csv, index=False)
                    logger.info("Writing issues out to file %s", val_csv)
                    sys.exit(0)
                else:
                    # user may be in python session, return dataframe
                    return parsed

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids:ro"
    output_dir_link = str(opts.output_prefix.parent.absolute()) + ":/csv:rw"
    linked_output_prefix = "/csv/" + opts.output_prefix.name
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm', '-v', bids_dir_link,
               '-v', GIT_CONFIG+":/root/.gitconfig",
               '-v', output_dir_link, '--entrypoint', 'cubids-validate',
               opts.container, '/bids', linked_output_prefix]
        if opts.ignore_nifti_headers:
            cmd.append('--ignore_nifti_headers')
        if opts.ignore_subject_consistency:
            cmd.append('--ignore_subject_consistency')
    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               '-B', output_dir_link, opts.container, 'cubids-validate',
               '/bids', linked_output_prefix]
        if opts.ignore_nifti_headers:
            cmd.append('--ignore_nifti_headers')
        if opts.ignore_subject_consistency:
            cmd.append('--ignore_subject_consistency')
        if opts.sequential:
            cmd.append('--sequential')

    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def bids_sidecar_merge():
    parser = argparse.ArgumentParser(
        description="bids-sidecar-merge: merge critical keys from one "
        "sidecar to another",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('from_json',
                        type=Path,
                        action='store',
                        help='Source json file.')
    parser.add_argument('to_json',
                        type=Path,
                        action='store',
                        help='destination json. This file will have data '
                        'from `from_json` copied into it.')
    opts = parser.parse_args()
    merge_status = merge_json_into_json(opts.from_json, opts.to_json,
                                        raise_on_error=False)
    sys.exit(merge_status)


def cubids_group():
    '''Command Line Interface function for finding key and param groups.'''

    parser = argparse.ArgumentParser(
        description="cubids-group: find key and parameter groups in BIDS",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='the root of a BIDS dataset. It should contain '
                        'sub-X directories and dataset_description.json')
    parser.add_argument('output_prefix',
                        type=Path,
                        action='store',
                        help='file prefix to which a _summary.csv, _files.csv '
                        'and _group.csv are written.')
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    parser.add_argument('--acq-group-level',
                        default='subject',
                        action='store',
                        help='Level at which acquisition groups are created '
                        'options: "subject" or "session"')
    parser.add_argument('--config',
                        action='store',
                        type=Path,
                        help='path to a config file for grouping')
    opts = parser.parse_args()

    # Run directly from python using
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir),
                     acq_group_level=opts.acq_group_level,
                     grouping_config=opts.config)
        bod.get_CSVs(str(opts.output_prefix),)
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids"
    output_dir_link = str(opts.output_prefix.parent.absolute()) + ":/csv:rw"

    apply_config = opts.config is not None
    if apply_config:
        input_config_dir_link = str(
            opts.config.parent.absolute()) + ":/in_config:ro"
        linked_input_config = "/in_config/" + opts.config.name

    linked_output_prefix = "/csv/" + opts.output_prefix.name
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm', '-v', bids_dir_link,
               '-v', GIT_CONFIG+":/root/.gitconfig",
               '-v', output_dir_link,
               '--entrypoint', 'cubids-group',
               opts.container, '/bids', linked_output_prefix]
        if apply_config:
            cmd.insert(3, '-v')
            cmd.insert(4, input_config_dir_link)
            cmd += ['--config', linked_input_config]

    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               '-B', output_dir_link,
               opts.container, 'cubids-group',
               '/bids', linked_output_prefix]
        if apply_config:
            cmd.insert(3, '-B')
            cmd.insert(4, input_config_dir_link)
            cmd += ['--config', linked_input_config]

    if opts.acq_group_level:
        cmd.append("--acq-group-level")
        cmd.append(str(opts.acq_group_level))

    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def cubids_apply():
    ''' Command Line Interface funciton for applying the csv changes.'''

    parser = argparse.ArgumentParser(
        description="cubids-apply: apply the changes specified in a csv "
        "to a BIDS directory",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='the root of a BIDS dataset. It should contain '
                        'sub-X directories and dataset_description.json')
    parser.add_argument('edited_summary_csv',
                        type=Path,
                        action='store',
                        help='the _summary.csv that has been edited in the '
                        'MergeInto and RenameKeyGroup columns.')
    parser.add_argument('files_csv',
                        type=Path,
                        action='store',
                        help='the _files.csv that the _summary.csv '
                        'corresponds to.')
    parser.add_argument('new_csv_prefix',
                        type=Path,
                        action='store',
                        help='file prefix for writing the new _summary.csv, '
                        '_files.csv and _group.csv that have been edited.')
    parser.add_argument('--use-datalad',
                        action='store_true',
                        help='ensure that there are no untracked changes '
                        'before finding groups')
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    parser.add_argument('--acq-group-level',
                        default='subject',
                        action='store',
                        help='Level at which acquisition groups are created '
                        'options: "subject" or "session"')
    parser.add_argument('--config',
                        action='store',
                        type=Path,
                        help='path to a config file for grouping')
    opts = parser.parse_args()

    # Run directly from python using
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir),
                     use_datalad=opts.use_datalad,
                     acq_group_level=opts.acq_group_level,
                     grouping_config=opts.config)
        if opts.use_datalad:
            if not bod.is_datalad_clean():
                raise Exception("Untracked change in " + str(opts.bids_dir))
        bod.apply_csv_changes(str(opts.edited_summary_csv),
                              str(opts.files_csv),
                              str(opts.new_csv_prefix),
                              raise_on_error=False)
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids"
    input_summary_csv_dir_link = str(
        opts.edited_csv_prefix.parent.absolute()) + ":/in_summary_csv:ro"
    input_files_csv_dir_link = str(
        opts.edited_csv_prefix.parent.absolute()) + ":/in_files_csv:ro"
    output_csv_dir_link = str(
        opts.new_csv_prefix.parent.absolute()) + ":/out_csv:rw"

    # FROM BOND-GROUP
    apply_config = opts.config is not None
    if apply_config:
        input_config_dir_link = str(
            opts.config.parent.absolute()) + ":/in_config:ro"
        linked_input_config = "/in_config/" + opts.config.name

    linked_output_prefix = "/csv/" + opts.output_prefix.name

    ####

    linked_input_summary_csv = "/in_summary_csv/" \
        + opts.edited_summary_csv.name
    linked_input_files_csv = "/in_files_csv/" + opts.files_csv.name
    linked_output_prefix = "/out_csv/" + opts.new_csv_prefix.name
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm',
               '-v', bids_dir_link,
               '-v', GIT_CONFIG+":/root/.gitconfig",
               '-v', input_summary_csv_dir_link,
               '-v', input_files_csv_dir_link,
               '-v', output_csv_dir_link,
               '--entrypoint', 'cubids-apply',
               opts.container, '/bids', linked_input_summary_csv,
               linked_input_files_csv, linked_output_prefix]
        if apply_config:
            cmd.insert(3, '-v')
            cmd.insert(4, input_config_dir_link)
            cmd += ['--config', linked_input_config]

    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               '-B', input_summary_csv_dir_link,
               '-B', input_files_csv_dir_link,
               '-B', output_csv_dir_link,
               opts.container, 'cubids-apply',
               '/bids', linked_input_summary_csv,
               linked_input_files_csv, linked_output_prefix]
        if apply_config:
            cmd.insert(3, '-B')
            cmd.insert(4, input_config_dir_link)
            cmd += ['--config', linked_input_config]

    if opts.use_datalad:
        cmd.append("--use-datalad")

    if opts.acq_group_level:
        cmd.append("--acq-group-level")
        cmd.append(str(opts.acq_group_level))

    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def cubids_datalad_save():
    ''' Command Line Interfcae function for performing datalad save.'''

    parser = argparse.ArgumentParser(
        description="cubids-datalad-save: perform a DataLad save on a BIDS "
        "directory",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='the root of a BIDS dataset. It should contain '
                        'sub-X directories and dataset_description.json')
    parser.add_argument('-m',
                        action='store',
                        help='message for this commit')
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    opts = parser.parse_args()

    # Run directly from python using
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir), use_datalad=True)
        bod.datalad_save(message=opts.m)
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids"
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm', '-v', bids_dir_link,
               '-v', GIT_CONFIG+":/root/.gitconfig",
               '--entrypoint', 'cubids-datalad-save',
               opts.container, '/bids', '-m', opts.m]
    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               opts.container, 'cubids-datalad-save',
               '/bids', '-m', opts.m]
    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def cubids_undo():
    ''' Command Line Interface function for reverting a commit.'''

    parser = argparse.ArgumentParser(
        description="cubids-undo: revert most recent commit",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='the root of a BIDS dataset. It should contain '
                        'sub-X directories and dataset_description.json')
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    opts = parser.parse_args()

    # Run directly from python using
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir), use_datalad=True)
        bod.datalad_undo_last_commit()
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids"
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm', '-v', bids_dir_link,
               '-v', GIT_CONFIG+":/root/.gitconfig",
               '--entrypoint', 'cubids-undo',
               opts.container, '/bids']
    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               opts.container, 'cubids-undo', '/bids']
    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def cubids_copy_exemplars():
    ''' Command Line Interface function for purging scan associations.'''

    parser = argparse.ArgumentParser(
        description="cubids-copy-exemplars: create and save a directory with "
        " one subject from each Acquisition Group in the BIDS dataset",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='absolute path to the root of a BIDS dataset. '
                        'It should contain sub-X directories and '
                        'dataset_description.json.')
    parser.add_argument('exemplars_dir',
                        type=Path,
                        action='store',
                        help='absolute path to the root of a BIDS dataset '
                        'containing one subject from each Acquisition Group. '
                        'It should contain sub-X directories and '
                        'dataset_description.json.')
    parser.add_argument('exemplars_csv',
                        type=Path,
                        action='store',
                        help='absolute path to the .csv file that lists one '
                        'subject from each Acqusition Group '
                        '(*_AcqGrouping.csv from the cubids-group output)')
    parser.add_argument('--use-datalad',
                        action='store_true',
                        help='check exemplar dataset into DataLad')
    parser.add_argument('--min-group-size',
                        action='store',
                        default=1,
                        help='minimum number of subjects an Acquisition Group '
                        'must have in order to be included in the exemplar '
                        'dataset ',
                        required=False)
    # parser.add_argument('--include-groups',
    #                     action='store',
    #                     nargs='+',
    #                     default=[],
    #                     help='only include an exemplar subject from these '
    #                     'listed Acquisition Groups in the exemplar dataset ',
    #                     required=False)
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    opts = parser.parse_args()

    # Run directly from python using
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir),
                     use_datalad=opts.use_datalad)
        if opts.use_datalad:
            if not bod.is_datalad_clean():
                raise Exception("Untracked changes. Need to save "
                                + str(opts.bids_dir) +
                                " before coyping exemplars")
        bod.copy_exemplars(str(opts.exemplars_dir), str(opts.exemplars_csv),
                           min_group_size=opts.min_group_size,
                           raise_on_error=True)
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids:ro"
    exemplars_dir_link = str(opts.exemplars_dir.absolute()) + ":/exemplars:ro"
    exemplars_csv_link = str(opts.exemplars_csv.absolute()) + ":/in_csv:ro"
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm', '-v', bids_dir_link,
               '-v', exemplars_dir_link,
               '-v', GIT_CONFIG+":/root/.gitconfig",
               '-v', exemplars_csv_link, '--entrypoint',
               'cubids-copy-exemplars',
               opts.container, '/bids', '/exemplars', '/in_csv']

        if opts.force_unlock:
            cmd.append('--force-unlock')
        if opts.min_group_size:
            cmd.append('--min-group-size')
    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               '-B', exemplars_dir_link,
               '-B', exemplars_csv_link, opts.container,
               'cubids-copy-exemplars',
               '/bids', '/exemplars', '/in_csv']
        if opts.force_unlock:
            cmd.append('--force-unlock')
        if opts.min_group_size:
            cmd.append('--min-group-size')

    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def cubids_add_nifti_info():
    ''' Command Line Interface function for purging scan associations.'''

    parser = argparse.ArgumentParser(
        description="cubids-add-nifti-info: Add information from nifti"
        "files to the sidecars of each dataset",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='absolute path to the root of a BIDS dataset. '
                        'It should contain sub-X directories and '
                        'dataset_description.json.')
    parser.add_argument('--use-datalad',
                        action='store_true',
                        help='ensure that there are no untracked changes '
                        'before finding groups')
    parser.add_argument('--force-unlock',
                        action='store_true',
                        help='unlock dataset before adding nifti info ')
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    opts = parser.parse_args()

    # Run directly from python using
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir),
                     use_datalad=opts.use_datalad,
                     force_unlock=opts.force_unlock)
        if opts.use_datalad:
            if not bod.is_datalad_clean():
                raise Exception("Untracked change in " + str(opts.bids_dir))
            # if bod.is_datalad_clean() and not opts.force_unlock:
            #     raise Exception("Need to unlock " + str(opts.bids_dir))
        bod.add_nifti_info()
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids:ro"
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm', '-v', bids_dir_link,
               '-v', GIT_CONFIG+":/root/.gitconfig",
               '--entrypoint', 'cubids-add-nifti-info',
               opts.container, '/bids']

        if opts.force_unlock:
            cmd.append('--force-unlock')
    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               opts.container, 'cubids-add-nifti-info',
               '/bids']
        if opts.force_unlock:
            cmd.append('--force-unlock')

    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def cubids_purge():
    ''' Command Line Interface function for purging scan associations.'''

    parser = argparse.ArgumentParser(
        description="cubids-purge: purge associations from the dataset",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='absolute path to the root of a BIDS dataset. '
                        'It should contain sub-X directories and '
                        'dataset_description.json.')
    parser.add_argument('scans',
                        type=Path,
                        action='store',
                        help='absolute path to the txt file of scans whose '
                        'associations should be purged.')
    parser.add_argument('--use-datalad',
                        action='store_true',
                        help='ensure that there are no untracked changes '
                        'before finding groups')
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    opts = parser.parse_args()

    # Run directly from python using
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir),
                     use_datalad=opts.use_datalad)
        if opts.use_datalad:
            if not bod.is_datalad_clean():
                raise Exception("Untracked change in " + str(opts.bids_dir))
        bod.purge(str(opts.scans), raise_on_error=False)
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids"
    input_scans_link = str(
        opts.scans.parent.absolute()) + ":/in_scans:ro"
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm',
               '-v', bids_dir_link,
               '-v', GIT_CONFIG+":/root/.gitconfig",
               '-v', input_scans_link,
               '--entrypoint', 'cubids-purge',
               opts.container, '/bids', input_scans_link]

    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               '-B', input_scans_link,
               opts.container, 'cubids-purge',
               '/bids', input_scans_link]
    print("RUNNING: " + ' '.join(cmd))
    if opts.use_datalad:
        cmd.append("--use-datalad")
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def cubids_remove_metadata_fields():
    ''' Command Line Interface function for deteling fields from metadata.'''

    parser = argparse.ArgumentParser(
        description="cubids-remove-metadata-fields: delete fields from "
        "metadata",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='the root of a BIDS dataset. It should contain '
                        'sub-X directories and dataset_description.json')
    parser.add_argument('--fields',
                        nargs='+',
                        action='store',
                        default=[],
                        help='space-separated list of metadata fields to '
                        'remove.')
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    opts = parser.parse_args()

    # Run directly from python
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir), use_datalad=False)
        bod.remove_metadata_fields(opts.fields)
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids:rw"
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm', '-v', bids_dir_link,
               '--entrypoint', 'cubids-remove-metadata-fields',
               opts.container, '/bids', '--fields'] + opts.fields
    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               opts.container, 'cubids-remove-metadata-fields',
               '/bids', '--fields'] + opts.fields
    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def cubids_print_metadata_fields():
    '''Command Line Interface function that prints unique metadata fields.'''

    parser = argparse.ArgumentParser(
        description="cubids-print-metadata-fields: print all unique "
        "metadata fields",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bids_dir',
                        type=Path,
                        action='store',
                        help='the root of a BIDS dataset. It should contain '
                        'sub-X directories and dataset_description.json')
    parser.add_argument('--container',
                        action='store',
                        help='Docker image tag or Singularity image file.')
    opts = parser.parse_args()

    # Run directly from python
    if opts.container is None:
        bod = CuBIDS(data_root=str(opts.bids_dir), use_datalad=False)
        fields = bod.get_all_metadata_fields()
        print("\n".join(fields))
        sys.exit(0)

    # Run it through a container
    container_type = _get_container_type(opts.container)
    bids_dir_link = str(opts.bids_dir.absolute()) + ":/bids:ro"
    if container_type == 'docker':
        cmd = ['docker', 'run', '--rm', '-v', bids_dir_link,
               '--entrypoint', 'cubids-print-metadata-fields',
               opts.container, '/bids']
    elif container_type == 'singularity':
        cmd = ['singularity', 'exec', '--cleanenv',
               '-B', bids_dir_link,
               opts.container, 'cubids-print-metadata-fields',
               '/bids']
    print("RUNNING: " + ' '.join(cmd))
    proc = subprocess.run(cmd)
    sys.exit(proc.returncode)


def _get_container_type(image_name):
    '''Gets and returns the container type.'''

    # If it's a file on disk, it must be a singularity image
    if Path(image_name).exists():
        return "singularity"

    # It needs to match a docker tag pattern to be docker
    if re.match(r"(?:.+\/)?([^:]+)(?::.+)?", image_name):
        return "docker"

    raise Exception("Unable to determine the container type of "
                    + image_name)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

# with open('README.rst') as readme_file:
#     readme = readme_file.read()

with open("README.rst", "r", encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

setup_requirements = ['pytest-runner']

setup(
    author="Neuroinformatics Team of PennLINC",
    author_email='sydney.covitz@pennmecidine.upenn.edu',
    maintainer='Sydney Covitz',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="BIDS Curation Tool",
    entry_points={
        'console_scripts': [
            'cubids-group=cubids.cli:cubids_group',
            'cubids-apply=cubids.cli:cubids_apply',
            'cubids-purge=cubids.cli:cubids_purge',
            'cubids-add-nifti-info=cubids.cli:cubids_add_nifti_info',
            'cubids-copy-exemplars=cubids.cli:cubids_copy_exemplars',
            'cubids-undo=cubids.cli:cubids_undo',
            'bids-sidecar-merge=cubids.cli:bids_sidecar_merge',
            'cubids-validate=cubids.cli:cubids_validate',
            'cubids-datalad-save=cubids.cli:cubids_datalad_save',
            'cubids-print-metadata-fields=cubids.cli:'
            'cubids_print_metadata_fields',
            'cubids-remove-metadata-fields=cubids.cli:'
            'cubids_remove_metadata_fields'
        ],
    },
    license="MIT License",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='cubids',
    name='cubids',
    packages=find_packages(include=['cubids', 'cubids.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    url='https://github.com/pennlinc/cubids',
    version='1.0.2',
    zip_safe=False,
)
